from .builders import (
    TypeHintBuilder,
    build_union,
)
from .comparators import (
    TypeHintComparator,
    is_model_subclass,
    is_model_type,
    is_type_subclass,
)
from .constants import (
    MissingSentinel,
    NoneType,
)
from .generics import (
    GenericTypeProjector,
    unpack_typevar,
)
from .model_types import (
    FieldType,
    ModelType,
)



import abc
from typing import (
    Any,
)


class MinosBinaryProtocol(abc.ABC):
    """Minos binary encoder / decoder interface."""

    @classmethod
    @abc.abstractmethod
    def encode(cls, *args, **kwargs) -> bytes:
        """Encodes the given value into bytes.

        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: A bytes instance.
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def decode(cls, data: bytes, *args, **kwargs) -> Any:
        """Decodes the given bytes data.

        :param data: bytes data to be decoded.
        :return: De decoded data.
        """
        raise NotImplementedError



from __future__ import (
    annotations,
)

import logging
import warnings
from asyncio import (
    AbstractEventLoop,
    gather,
)
from enum import (
    Enum,
)
from types import (
    ModuleType,
)
from typing import (
    NoReturn,
    Optional,
    Union,
)

from aiomisc.entrypoint import (
    Entrypoint,
)
from aiomisc.log import (
    LogFormat,
    basic_config,
)
from aiomisc.utils import (
    create_default_event_loop,
)
from aiomisc_log.enum import (
    DateFormat,
)
from cached_property import (
    cached_property,
)

from .config import (
    Config,
)
from .importlib import (
    get_internal_modules,
    import_module,
)
from .injections import (
    DependencyInjector,
    InjectableMixin,
)
from .ports import (
    Port,
)
from .setup import (
    SetupMixin,
)

logger = logging.getLogger(__name__)


def _create_entrypoint(*args, **kwargs) -> Entrypoint:  # pragma: no cover
    return Entrypoint(*args, **kwargs)


def _create_loop() -> AbstractEventLoop:  # pragma: no cover
    return create_default_event_loop()[0]


class EntrypointLauncher(SetupMixin):
    """EntryPoint Launcher class."""

    def __init__(
        self,
        config: Config,
        injections: list[Union[SetupMixin, type[SetupMixin], str]],
        ports: list[Union[Port, type[Port], str]],
        log_level: Union[int, str] = logging.INFO,
        log_format: Union[str, LogFormat] = "color",
        log_date_format: Union[str, DateFormat] = DateFormat["color"],
        external_modules: Optional[list[ModuleType]] = None,
        external_packages: Optional[list[str]] = None,
        *args,
        **kwargs,
    ):
        if external_modules is None:
            external_modules = list()
        if external_packages is None:
            external_packages = list()

        super().__init__(*args, **kwargs)

        if isinstance(log_date_format, Enum):
            log_date_format = log_date_format.value

        self.config = config

        self._log_level = log_level
        self._log_format = log_format
        self._log_date_format = log_date_format

        self._raw_injections = injections
        self._raw_ports = ports
        self._external_modules = external_modules
        self._external_packages = external_packages

    @classmethod
    def _from_config(cls, *args, config: Config, **kwargs) -> EntrypointLauncher:
        if "injections" not in kwargs:
            kwargs["injections"] = config.get_injections()
        if "ports" not in kwargs:
            kwargs["ports"] = [
                interface["port"] for interface in config.get_interfaces().values() if "port" in interface
            ]
        return cls(config, *args, **kwargs)

    def launch(self) -> NoReturn:
        """Launch a new execution and keeps running forever..

        :return: This method does not return anything.
        """

        basic_config(
            level=self._log_level, log_format=self._log_format, buffered=False, date_format=self._log_date_format
        )

        logger.info("Starting microservice...")

        exception = None
        try:
            self.graceful_launch()
            logger.info("Microservice is up and running!")
            self.loop.run_forever()
        except KeyboardInterrupt as exc:  # pragma: no cover
            logger.info("Stopping microservice...")
            exception = exc
        except Exception as exc:  # pragma: no cover
            logger.exception("Stopping microservice due to an unhandled exception...")
            exception = exc
        finally:
            self.graceful_shutdown(exception)

    def graceful_launch(self) -> None:
        """Launch the execution gracefully.

        :return: This method does not return anything.
        """
        self.loop.run_until_complete(self.setup())

    def graceful_shutdown(self, err: Exception = None) -> None:
        """Shutdown the execution gracefully.

        :return: This method does not return anything.
        """
        self.loop.run_until_complete(self.destroy())

    @cached_property
    def entrypoint(self) -> Entrypoint:
        """Entrypoint instance.

        :return: An ``Entrypoint`` instance.
        """
        return _create_entrypoint(*self.ports, loop=self.loop, log_config=False)

    @cached_property
    def loop(self) -> AbstractEventLoop:
        """Create the loop.

        :return: An ``AbstractEventLoop`` instance.
        """
        return _create_loop()

    @cached_property
    def ports(self) -> list[Port]:
        """List of ports to be launched.

        :return: A list of ``Port`` instances.
        """

        def _fn(raw: Union[Port, type[Port], str]) -> Port:
            if isinstance(raw, str):
                raw = import_module(raw)
            if isinstance(raw, type):
                return raw(config=self.config)
            return raw

        return [_fn(raw) for raw in self._raw_ports]

    @property
    def services(self) -> list[Port]:
        """List of ports to be launched.

        :return: A list of ``Port`` instances.
        """
        warnings.warn("'services' property has been deprecated. Use 'ports' instead.", DeprecationWarning)

        return self.ports

    async def _setup(self) -> None:
        """Wire the dependencies and setup it.

        :return: This method does not return anything.
        """
        modules = self._external_modules + self._internal_modules
        packages = self._external_packages
        self.injector.wire_injections(modules=modules, packages=packages)
        await gather(self.injector.setup_injections(), self.entrypoint.__aenter__())

    @property
    def _internal_modules(self) -> list[ModuleType]:
        return get_internal_modules()

    async def _destroy(self) -> None:
        """Unwire the injected dependencies and destroys it.

        :return: This method does not return anything.
        """
        await gather(self.entrypoint.__aexit__(None, None, None), self.injector.destroy_injections())
        self.injector.unwire_injections()

    @property
    def injections(self) -> dict[str, InjectableMixin]:
        """Get the injections mapping.

        :return: A ``dict`` with injection names as keys and injection instances as values.
        """
        return self.injector.injections

    @cached_property
    def injector(self) -> DependencyInjector:
        """Dependency injector instance.

        :return: A ``DependencyInjector`` instance.
        """
        return DependencyInjector(self.config, self._raw_injections)



from __future__ import (
    annotations,
)

from ..types import (
    ModelType,
)
from .abc import (
    DynamicModel,
)


class DataTransferObject(DynamicModel):
    """Data Transfer Object to build the objects dynamically from bytes"""

    def __init__(self, name: str, *args, namespace: str = str(), **kwargs):
        super().__init__(*args, **kwargs)
        if "." in name:
            raise ValueError(
                "The 'name' attribute cannot contain dots. You can use the 'namespace' to qualify the instance"
            )
        self._name = name
        self._namespace = namespace

    @classmethod
    def from_model_type(cls, model_type: ModelType, *args, **kwargs) -> DataTransferObject:
        """Build a ``DataTransferObject`` from a ``ModelType``.

        :param model_type: ``ModelType`` object containing the model structure
        :param args: Positional arguments to be passed to the model constructor.
        :param kwargs: Named arguments to be passed to the model constructor.
        :return: A new ``DataTransferObject`` instance.
        """
        fields = cls._build_fields(model_type.type_hints, *args, **kwargs)
        return cls(model_type.name, fields, namespace=model_type.namespace)

    @property
    def classname(self) -> str:
        """Compute the current class name.

        :return: An string object.
        """
        if len(self._namespace) == 0:
            return self._name
        return f"{self._namespace}.{self._name}"

    def __repr__(self) -> str:
        fields_repr = ", ".join(repr(field) for field in self.fields.values())
        return f"{self._name}[DTO]({fields_repr})"

    def __eq__(self, other: DataTransferObject):
        return super().__eq__(other) and self._name == other._name and self._namespace == other._namespace



from ..setup import (
    SetupMixin,
)


class InjectableMixin(SetupMixin):
    """Injectable Mixin class."""

    _injectable_name: str

    @classmethod
    def _set_injectable_name(cls, name: str) -> None:
        cls._injectable_name = name

    @classmethod
    def get_injectable_name(cls) -> str:
        """Get the injectable name.

        :return: A ``str`` value.
        """
        return cls._injectable_name



from __future__ import (
    annotations,
)

from contextlib import (
    suppress,
)
from typing import (
    TYPE_CHECKING,
    Any,
)

from ..exceptions import (
    MinosConfigException,
)
from ..importlib import (
    import_module,
)
from .abc import (
    Config,
)

if TYPE_CHECKING:
    from ..injections import (
        InjectableMixin,
    )


class ConfigV1(Config):
    """
    A Minos configuration provides information on the connection points available at that service.
    It consists of the following parts:

    - Service meta-information (such as name, or version).
    - REST Service endpoints available.
    - Repository database connection for event sourcing.
    - Snapshot database connection.
    - Events it publishes/consumes from de given Kafka service.
    - Commands it reacts to from other microservices.
    - Sagas it takes part on.
    """

    _ENVIRONMENT_MAPPER = {
        "service.name": "MINOS_SERVICE_NAME",
        "rest.host": "MINOS_REST_HOST",
        "rest.port": "MINOS_REST_PORT",
        "broker.host": "MINOS_BROKER_HOST",
        "broker.port": "MINOS_BROKER_PORT",
        "broker.queue.host": "MINOS_BROKER_QUEUE_HOST",
        "broker.queue.port": "MINOS_BROKER_QUEUE_PORT",
        "broker.queue.database": "MINOS_BROKER_QUEUE_DATABASE",
        "broker.queue.user": "MINOS_BROKER_QUEUE_USER",
        "broker.queue.password": "MINOS_BROKER_QUEUE_PASSWORD",
        "commands.service": "MINOS_COMMANDS_SERVICE",
        "queries.service": "MINOS_QUERIES_SERVICE",
        "repository.host": "MINOS_REPOSITORY_HOST",
        "repository.port": "MINOS_REPOSITORY_PORT",
        "repository.database": "MINOS_REPOSITORY_DATABASE",
        "repository.user": "MINOS_REPOSITORY_USER",
        "repository.password": "MINOS_REPOSITORY_PASSWORD",
        "snapshot.host": "MINOS_SNAPSHOT_HOST",
        "snapshot.port": "MINOS_SNAPSHOT_PORT",
        "snapshot.database": "MINOS_SNAPSHOT_DATABASE",
        "snapshot.user": "MINOS_SNAPSHOT_USER",
        "snapshot.password": "MINOS_SNAPSHOT_PASSWORD",
        "discovery.client": "MINOS_DISCOVERY_CLIENT",
        "discovery.host": "MINOS_DISCOVERY_HOST",
        "discovery.port": "MINOS_DISCOVERY_PORT",
    }

    _PARAMETERIZED_MAPPER = {
        "service.name": "service_name",
        "service.injections": "service_injections",
        "rest.host": "rest_host",
        "rest.port": "rest_port",
        "broker.host": "broker_host",
        "broker.port": "broker_port",
        "broker.queue.host": "broker_queue_host",
        "broker.queue.port": "broker_queue_port",
        "broker.queue.database": "broker_queue_database",
        "broker.queue.user": "broker_queue_user",
        "broker.queue.password": "broker_queue_password",
        "commands.service": "commands_service",
        "queries.service": "queries_service",
        "saga.broker": "saga_broker",
        "saga.port": "saga_port",
        "repository.host": "repository_host",
        "repository.port": "repository_port",
        "repository.database": "repository_database",
        "repository.user": "repository_user",
        "repository.password": "repository_password",
        "snapshot.host": "snapshot_host",
        "snapshot.port": "snapshot_port",
        "snapshot.database": "snapshot_database",
        "snapshot.user": "snapshot_user",
        "snapshot.password": "snapshot_password",
        "discovery.client": "minos_discovery_client",
        "discovery.host": "minos_discovery_host",
        "discovery.port": "minos_discovery_port",
    }

    @property
    def _version(self) -> int:
        return 1

    def _get_name(self) -> str:
        return self.get_by_key("service.name")

    def _get_aggregate(self) -> dict[str, Any]:
        return {
            "entities": [self.get_type_by_key("service.aggregate")],
            "repositories": dict(),
        }

    def _get_saga(self) -> dict[str, Any]:
        try:
            saga = self.get_by_key("saga")
        except MinosConfigException:
            saga = dict()
        saga.pop("storage", None)
        return saga

    def _get_injections(self) -> list[type[InjectableMixin]]:
        try:
            injections = self.get_by_key("service.injections")
            if isinstance(injections, dict):
                injections = list(injections.values())
        except MinosConfigException:
            injections = list()

        injections = [import_module(classname) for classname in injections]

        # noinspection PyTypeChecker
        return injections

    def _get_interfaces(self) -> dict[str, dict[str, Any]]:
        interfaces = dict()

        with suppress(MinosConfigException):
            interfaces["http"] = self._get_interface_http()

        with suppress(MinosConfigException):
            interfaces["broker"] = self._get_interface_broker()

        with suppress(MinosConfigException):
            interfaces["periodic"] = self._get_interface_periodic()

        return interfaces

    def _get_interface_http(self) -> dict[str, Any]:
        try:
            port = next(
                port
                for port in self.get_by_key("service.services")
                if ("http" in port.lower() or "rest" in port.lower())
            )
        except Exception as exc:
            raise MinosConfigException(f"The 'http' interface is not available: {exc!r}")

        try:
            connector = self.get_by_key("rest")
        except MinosConfigException:
            connector = dict()

        return {
            "port": import_module(port),
            "connector": connector,
        }

    def _get_interface_broker(self) -> dict[str, Any]:
        try:
            port = next(port for port in self.get_by_key("service.services") if "broker" in port.lower())
        except Exception as exc:
            raise MinosConfigException(f"The 'broker' interface is not available: {exc!r}")

        try:
            common = self.get_by_key("broker")
        except MinosConfigException:
            common = dict()

        try:
            common["queue"] = self.get_by_key("broker.queue")
            common["queue"].pop("database", None)
            common["queue"].pop("port", None)
            common["queue"].pop("host", None)
            common["queue"].pop("port", None)
            common["queue"].pop("user", None)
            common["queue"].pop("password", None)
        except MinosConfigException:
            common["queue"] = dict()

        return {
            "port": import_module(port),
            "publisher": dict(),
            "subscriber": dict(),
            "common": common,
        }

    def _get_interface_periodic(self):
        try:
            port = next(port for port in self.get_by_key("service.services") if "periodic" in port.lower())
        except Exception as exc:
            raise MinosConfigException(f"The 'periodic' interface is not available: {exc!r}")

        return {
            "port": import_module(port),
        }

    def _get_services(self) -> list[type]:
        try:
            services = self.get_by_key("services")
        except MinosConfigException:
            services = list()

        services = [import_module(classname) for classname in services]

        return services

    def _get_pools(self) -> dict[str, type]:
        return dict()

    def _get_routers(self) -> list[type]:
        try:
            routers = self.get_by_key("routers")
        except MinosConfigException:
            routers = list()

        routers = [import_module(classname) for classname in routers]

        return routers

    def _get_middleware(self) -> list[type]:
        try:
            middleware = self.get_by_key("middleware")
        except MinosConfigException:
            middleware = list()

        middleware = [import_module(classname) for classname in middleware]

        return middleware

    def _get_databases(self) -> dict[str, dict[str, Any]]:
        databases = dict()

        with suppress(MinosConfigException):
            databases["broker"] = self._get_database_broker()

        with suppress(MinosConfigException):
            databases["event"] = self._get_database_event()

        with suppress(MinosConfigException):
            databases["snapshot"] = self._get_database_snapshot()

        with suppress(MinosConfigException):
            databases["saga"] = self._get_database_saga()

        with suppress(MinosConfigException):
            databases["query"] = self._get_database_query()

        with suppress(MinosConfigException):
            databases["default"] = self._get_database_event()

        return databases

    def _get_database_broker(self):
        return self._get_database_by_name("broker.queue")

    def _get_database_saga(self) -> dict[str, Any]:
        raw = self.get_by_key("saga.storage")
        return raw

    def _get_database_event(self) -> dict[str, Any]:
        return self._get_database_by_name("repository")

    def _get_database_query(self) -> dict[str, Any]:
        return self._get_database_by_name("query_repository")

    def _get_database_snapshot(self) -> dict[str, Any]:
        return self._get_database_by_name("snapshot")

    def _get_database_by_name(self, prefix: str):
        data = self.get_by_key(prefix)
        data.pop("records", None)
        data.pop("retry", None)
        return data

    def _get_discovery(self) -> dict[str, Any]:
        data = self.get_by_key("discovery")
        data["client"] = self.get_type_by_key("discovery.client")
        return data

    def _to_parameterized_variable(self, key: str) -> str:
        return self._PARAMETERIZED_MAPPER[key]

    def _to_environment_variable(self, key: str) -> str:
        return self._ENVIRONMENT_MAPPER[key]



from __future__ import (
    annotations,
)

import logging
from datetime import (
    date,
    datetime,
    time,
    timedelta,
)
from enum import (
    Enum,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Type,
    Union,
    get_args,
    get_origin,
)
from uuid import (
    UUID,
    uuid4,
)

from .....importlib import (
    classname,
)
from ....types import (
    FieldType,
    MissingSentinel,
    ModelType,
    NoneType,
    is_model_subclass,
    is_type_subclass,
)
from ...abc import (
    SchemaEncoder,
)
from .constants import (
    AVRO_ARRAY,
    AVRO_BOOLEAN,
    AVRO_BYTES,
    AVRO_DATE,
    AVRO_DOUBLE,
    AVRO_INT,
    AVRO_MAP,
    AVRO_NULL,
    AVRO_SET,
    AVRO_STRING,
    AVRO_TIME,
    AVRO_TIMEDELTA,
    AVRO_TIMESTAMP,
    AVRO_UUID,
)

if TYPE_CHECKING:
    from ....abc import (
        Model,
    )
    from ....fields import (
        Field,
    )

logger = logging.getLogger(__name__)


class AvroSchemaEncoder(SchemaEncoder):
    """Avro Schema Encoder class."""

    def __init__(self, type_: type = None):
        self.type_ = type_

    def build(self, type_=MissingSentinel, **kwargs) -> Union[dict, list, str]:
        """Build the avro schema for the given field.

        :param type_: The type to be encoded as a schema.
        :return: A dictionary object.
        """
        if type_ is MissingSentinel:
            type_ = self.type_
        return self._build(type_, **kwargs)

    def _build(self, type_, **kwargs) -> Any:
        if get_origin(type_) is Union:
            return self._build_union(type_, **kwargs)

        if is_type_subclass(type_) and issubclass(type_, Enum):
            return self._build_enum(type_, **kwargs)

        return self._build_single(type_, **kwargs)

    def _build_union(self, type_: type, **kwargs) -> Any:
        ans = list()
        alternatives = get_args(type_)
        for alternative_type in alternatives:
            step = self._build(alternative_type, **kwargs)
            if isinstance(step, list):
                ans += step
            else:
                ans.append(step)
        return ans

    def _build_single(self, type_: type, **kwargs) -> Any:
        if type_ is Any:
            # FIXME: This is a design decision that must be revisited in the future.
            return AVRO_NULL

        if is_type_subclass(type_):
            if issubclass(type_, NoneType):
                return AVRO_NULL

            if issubclass(type_, bool):
                return AVRO_BOOLEAN

            if issubclass(type_, int):
                return AVRO_INT

            if issubclass(type_, float):
                return AVRO_DOUBLE

            if issubclass(type_, str):
                return AVRO_STRING

            if issubclass(type_, bytes):
                return AVRO_BYTES

            if issubclass(type_, datetime):
                return AVRO_TIMESTAMP

            if issubclass(type_, timedelta):
                return AVRO_TIMEDELTA

            if issubclass(type_, date):
                return AVRO_DATE

            if issubclass(type_, time):
                return AVRO_TIME

            if issubclass(type_, UUID):
                return AVRO_UUID

            if isinstance(type_, ModelType):
                return self._build_model_type(type_, **kwargs)

        from ....abc import (
            Model,
        )

        if isinstance(type_, Model) or is_model_subclass(type_):
            # noinspection PyTypeChecker
            return self._build_model(type_, **kwargs)

        from ....fields import (
            Field,
        )

        if isinstance(type_, Field):
            return self._build_field(type_, **kwargs)

        return self._build_collection(type_, **kwargs)

    def _build_enum(self, type_: type, **kwargs):
        return {"type": self._build_single(type_, **kwargs), "logicalType": classname(type_)}

    def _build_model(self, type_: Type[Model], **kwargs) -> Any:
        raw = ModelType.from_model(type_)
        return [self._build_model_type(raw, **kwargs)]

    def _build_model_type(self, type_: ModelType, **kwargs) -> Any:
        if (ans := type_.model_cls.encode_schema(self, type_, **kwargs)) is not MissingSentinel:
            return ans

        schema = {
            "name": type_.name,
            "namespace": self._patch_namespace(type_.namespace),
            "type": "record",
            "fields": [self._build_field(FieldType(n, t), **kwargs) for n, t in type_.type_hints.items()],
        }
        return schema

    @classmethod
    def _patch_namespace(cls, namespace: Optional[str]) -> Optional[str]:
        if len(namespace) > 0:
            namespace += f".{cls.generate_random_str()}"
        return namespace

    def _build_field(self, field: Union[Field, FieldType], **kwargs):
        return {"name": field.name, "type": self._build(field.type, **kwargs)}

    def _build_collection(self, type_: type, **kwargs) -> Any:
        origin_type = get_origin(type_)

        if origin_type is list:
            return self._build_list(type_, **kwargs)

        if origin_type is set:
            return self._build_set(type_, **kwargs)

        if origin_type is dict:
            return self._build_dict(type_, **kwargs)

        raise ValueError(f"Given field type is not supported: {type_}")  # pragma: no cover

    def _build_set(self, type_: type, **kwargs) -> dict[str, Any]:
        schema = self._build_list(type_, **kwargs)
        return schema | AVRO_SET

    def _build_list(self, type_: type, **kwargs) -> dict[str, Any]:
        return {"type": AVRO_ARRAY, "items": self._build(get_args(type_)[0], **kwargs)}

    def _build_dict(self, type_: type, **kwargs) -> dict[str, Any]:
        return {"type": AVRO_MAP, "values": self._build(get_args(type_)[1], **kwargs)}

    @staticmethod
    def generate_random_str() -> str:
        """Generate a random string

        :return: A random string value.
        """
        return str(uuid4())



from __future__ import (
    annotations,
)

from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
    Optional,
)


class MinosStorage(ABC):
    """Minos Storage interface."""

    @abstractmethod
    def add(self, **kwargs) -> None:
        """Store a value.

        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """
        raise NotImplementedError

    @abstractmethod
    def get(self, **kwargs) -> Optional[Any]:
        """Get the stored value..

        :param kwargs: Additional named arguments.
        :return: The stored value.
        """
        raise NotImplementedError

    @abstractmethod
    def delete(self, **kwargs) -> None:
        """Delete the stored value.

        :param kwargs:
        :return: This method does not return anything.
        """
        raise NotImplementedError

    @abstractmethod
    def update(self, **kwargs) -> None:
        """Update the stored value.

        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def build(cls, **kwargs) -> MinosStorage:
        """Build a new instance.

        :param kwargs: Additional named arguments.
        :return: A new ``MinosStorage`` instance.
        """
        raise NotImplementedError



from __future__ import (
    annotations,
)

from abc import (
    ABC,
)
from typing import (
    Any,
    Generic,
    Optional,
    TypeVar,
    Union,
    get_args,
)

from .config import (
    Config,
)
from .setup import (
    SetupMixin,
)

Instance = TypeVar("Instance")


class Builder(SetupMixin, ABC, Generic[Instance]):
    """Builder class."""

    def __init__(self, instance_cls: Optional[type[Instance]] = None, *args, **kwargs):

        super().__init__(*args, **kwargs)
        if instance_cls is None:
            instance_cls = self._get_cls()

        self.kwargs = dict()
        self.instance_cls = instance_cls

    def _get_cls(self) -> Optional[type]:
        # noinspection PyUnresolvedReferences
        bases = self.__orig_bases__

        instance_cls = get_args(next((base for base in bases if len(get_args(base))), None))[0]

        if not isinstance(instance_cls, type):
            return None

        return instance_cls

    def copy(self: type[B]) -> B:
        """Get a copy of the instance.

        :return: A ``Builder`` instance.
        """
        return self.new().with_cls(self.instance_cls).with_kwargs(self.kwargs)

    @classmethod
    def new(cls: type[B]) -> B:
        """Get a new instance.

        :return: A ``Builder`` instance.
        """
        return cls()

    def with_cls(self: B, cls: type) -> B:
        """Set class to be built.

        :param cls: The class to be set.
        :return: This method return the builder instance.
        """
        self.instance_cls = cls
        return self

    def with_kwargs(self: B, kwargs: dict[str, Any]) -> B:
        """Set kwargs.

        :param kwargs: The kwargs to be set.
        :return: This method return the builder instance.
        """
        self.kwargs |= kwargs
        return self

    # noinspection PyUnusedLocal
    def with_config(self: B, config: Config) -> B:
        """Set config.

        :param config: The config to be set.
        :return: This method return the builder instance.
        """
        return self

    def build(self) -> Instance:
        """Build the instance.

        :return: A ``Instance`` instance.
        """
        return self.instance_cls(**self.kwargs)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and self.instance_cls == other.instance_cls and self.kwargs == other.kwargs

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.instance_cls.__name__}, {self.kwargs!r})"


Ins = TypeVar("Ins", bound="BuildableMixin")


class BuildableMixin(SetupMixin):
    """Buildable Mixin class."""

    _builder: Union[Builder[Ins], type[Builder[Ins]]] = Builder

    @classmethod
    def _from_config(cls, config: Config, **kwargs):
        return cls.get_builder().with_config(config).with_kwargs(kwargs).build()

    @classmethod
    def set_builder(cls: type[Ins], builder: Union[Builder[Ins], type[Builder[Ins]]]) -> None:
        """Set a builder class.

        :param builder: The builder class to be set.
        :return: This method does not return anything.
        """
        if not isinstance(builder, Builder) and not (isinstance(builder, type) and issubclass(builder, Builder)):
            raise ValueError(f"Given builder value is invalid: {builder!r}")

        cls._builder = builder

    @classmethod
    def get_builder(cls) -> Builder[Ins]:
        """Get the builder class.

        :return: A ``Builder`` instance.
        """
        builder = cls._builder

        if isinstance(builder, Builder):
            builder = builder.copy()
        else:
            builder = builder.new()

        return builder.with_cls(cls)


B = TypeVar("B", bound=Builder)



import logging
from typing import (
    Any,
)

from .base import (
    MinosAvroProtocol,
)

logger = logging.getLogger(__name__)


class MinosAvroMessageProtocol(MinosAvroProtocol):
    """Minos Avro Messages Protocol class."""

    @classmethod
    def encode(cls, headers: dict, body: Any = None, *args, **kwargs) -> bytes:
        """
        encoder in avro
        all the headers are converted in fields with double underscore name
        the body is a set fields coming from the data type.
        """

        # prepare the headers
        final_data = dict()
        final_data["headers"] = headers
        if body:
            final_data["body"] = body

        return super().encode(final_data, _AVRO_SCHEMA)

    @classmethod
    def decode(cls, data: bytes, *args, **kwargs) -> dict:
        """Decode the given bytes of data into a single dictionary or a sequence of dictionaries.

        :param data: A bytes object.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: A dictionary or a list of dictionaries.
        """
        data_return = {"headers": dict()}
        for schema_dict in super().decode(data, batch_mode=True):
            logger.debug("Avro: get the request/response in dict format")
            data_return["headers"] = schema_dict["headers"]
            # check wich type is body
            if "body" in schema_dict:
                if isinstance(schema_dict["body"], dict):
                    data_return["body"] = {}
                elif isinstance(schema_dict["body"], list):
                    data_return["body"] = []
                data_return["body"] = schema_dict["body"]
            return data_return


_AVRO_SCHEMA = {
    "type": "record",
    "namespace": "org.minos.protocol",
    "name": "message",
    "fields": [
        {"name": "headers", "type": {"type": "map", "values": ["string", "int", "bytes", "long", "float", "boolean"]}},
        {
            "name": "body",
            "type": [
                "null",
                "string",
                "int",
                "bytes",
                "boolean",
                "float",
                "long",
                {"type": "array", "items": ["int", "string", "bytes", "long", "boolean", "float"]},
                {
                    "type": "map",
                    "values": [
                        "null",
                        "int",
                        "string",
                        "bytes",
                        "boolean",
                        "float",
                        "long",
                        {
                            "type": "array",
                            "items": [
                                "string",
                                "int",
                                "bytes",
                                "long",
                                "boolean",
                                "float",
                                {"type": "map", "values": ["string", "int", "bytes", "long", "boolean", "float"]},
                            ],
                        },
                        {
                            "type": "map",
                            "values": [
                                "string",
                                "int",
                                "bytes",
                                "long",
                                "boolean",
                                "float",
                                {"type": "map", "values": ["string", "int", "bytes", "long", "boolean", "float"]},
                            ],
                        },
                    ],
                },
            ],
        },
    ],
}



from __future__ import (
    annotations,
)

import logging
from abc import (
    abstractmethod,
)
from base64 import (
    b64decode,
    b64encode,
)
from collections.abc import (
    Mapping,
)
from typing import (
    Any,
    Iterable,
    Iterator,
    Type,
    TypedDict,
    TypeVar,
    Union,
)

from ..exceptions import (
    EmptyMinosModelSequenceException,
    MultiTypeMinosModelSequenceException,
)
from ..importlib import (
    classname,
)
from ..meta import (
    classproperty,
    property_or_classproperty,
    self_or_classmethod,
)
from ..protocol import (
    MinosAvroProtocol,
)
from .fields import (
    Field,
)
from .serializers import (
    AvroDataDecoder,
    AvroDataEncoder,
    AvroSchemaDecoder,
    AvroSchemaEncoder,
    DataDecoder,
    DataEncoder,
    SchemaDecoder,
    SchemaEncoder,
)
from .types import (
    MissingSentinel,
    ModelType,
)

logger = logging.getLogger(__name__)


class Model(Mapping):
    """Base class for ``minos`` model entities."""

    _field_cls: Type[Field] = Field

    _fields: dict[str, Field]
    __eq_reversing: bool

    def __init__(self, fields: Union[Iterable[Field], dict[str, Field]] = None, **kwargs):
        """Class constructor.

        :param fields: Dictionary that contains the ``Field`` instances of the model indexed by name.
        """
        if fields is None:
            fields = dict()
        if not isinstance(fields, dict):
            fields = {field.name: field for field in fields}
        self._fields = fields
        self.__eq_reversing = False

    @classmethod
    def from_typed_dict(cls: Type[T], typed_dict: TypedDict, *args, **kwargs) -> T:
        """Build a ``Model`` from a ``TypeDict`` and ``data``.

        :param typed_dict: ``TypeDict`` object containing the DTO's structure
        :param args: Positional arguments to be passed to the model constructor.
        :param kwargs: Named arguments to be passed to the model constructor.
        :return: A new ``DataTransferObject`` instance.
        """
        return cls.from_model_type(ModelType.from_typed_dict(typed_dict), *args, **kwargs)

    @classmethod
    @abstractmethod
    def from_model_type(cls: Type[T], model_type: ModelType, *args, **kwargs) -> T:
        """Build a ``Model`` from a ``ModelType``.

        :param model_type: ``ModelType`` object containing the DTO's structure
        :param args: Positional arguments to be passed to the model constructor.
        :param kwargs: Named arguments to be passed to the model constructor.
        :return: A new ``DeclarativeModel`` instance.
        """

    @classmethod
    def from_avro_str(cls: Type[T], raw: str, **kwargs) -> Union[T, list[T]]:
        """Build a single instance or a sequence of instances from bytes

        :param raw: A ``str`` representation of the model.
        :return: A single instance or a sequence of instances.
        """
        raw = b64decode(raw.encode())
        return cls.from_avro_bytes(raw, **kwargs)

    # noinspection PyUnusedLocal
    @classmethod
    def from_avro_bytes(cls: Type[T], raw: bytes, batch_mode: bool = False, **kwargs) -> Union[T, list[T]]:
        """Build a single instance or a sequence of instances from bytes.

        :param raw: A ``bytes`` representation of the model.
        :param batch_mode: If ``True`` the data is processed as a list of models, otherwise the data is processed as a
            single model.
        :param kwargs: Additional named arguments.
        :return: A single instance or a sequence of instances.
        """
        schema = MinosAvroProtocol.decode_schema(raw)
        data = MinosAvroProtocol.decode(raw, batch_mode=batch_mode)

        if batch_mode:
            return [cls.from_avro(schema, entry) for entry in data]

        return cls.from_avro(schema, data)

    @classmethod
    def from_avro(cls: Type[T], schema: Any, data: Any) -> T:
        """Build a new instance from the ``avro`` schema and data.

        :param schema: The avro schema of the model.
        :param data: The avro data of the model.
        :return: A new ``DynamicModel`` instance.
        """
        schema_decoder = AvroSchemaDecoder()
        type_ = schema_decoder.build(schema)

        data_decoder = AvroDataDecoder()
        instance = data_decoder.build(data, type_)

        return instance

    @classmethod
    def to_avro_str(cls: Type[T], models: list[T]) -> str:
        """Build the avro string representation of the given object instances.

        :param models: A sequence of minos models.
        :return: A bytes object.
        """
        return b64encode(cls.to_avro_bytes(models)).decode()

    @classmethod
    def to_avro_bytes(cls: Type[T], models: list[T]) -> bytes:
        """Create a ``bytes`` representation of the given object instances.

        :param models: A sequence of minos models.
        :return: A bytes object.
        """
        if len(models) == 0:
            raise EmptyMinosModelSequenceException("'models' parameter cannot be empty.")

        model_type = type(models[0])
        if not all(model_type == type(model) for model in models):
            raise MultiTypeMinosModelSequenceException(
                f"Every model must have type {model_type} to be valid. Found types: {[type(model) for model in models]}"
            )

        avro_schema = models[0].avro_schema
        # noinspection PyTypeChecker
        return MinosAvroProtocol.encode([model.avro_data for model in models], avro_schema, batch_mode=True)

    # noinspection PyMethodParameters
    @property_or_classproperty
    def model_type(self_or_cls) -> ModelType:
        """Get the model type of the instance.

        :return: A ``ModelType`` instance.
        """
        # noinspection PyTypeChecker
        return ModelType.from_model(self_or_cls)

    # noinspection PyMethodParameters
    @classproperty
    def classname(cls) -> str:
        """Compute the current class namespace.
        :return: An string object.
        """
        # noinspection PyTypeChecker
        return classname(cls)

    # noinspection PyMethodParameters
    @property_or_classproperty
    def type_hints(self_or_cls) -> dict[str, type]:
        """Get the type hinting of the instance or class.

        :return: A dictionary in which the keys are the field names and the values are the types.
        """
        return dict(self_or_cls._type_hints())

    # noinspection PyMethodParameters
    @self_or_classmethod
    def _type_hints(self_or_cls) -> Iterator[tuple[str, Any]]:
        if not isinstance(self_or_cls, type):
            yield from ((field.name, field.real_type) for field in self_or_cls.fields.values())

    # noinspection PyMethodParameters
    @property_or_classproperty
    def type_hints_parameters(self_or_cls) -> tuple[TypeVar, ...]:
        """Get the sequence of generic type hints parameters..

        :return: A tuple of `TypeVar` instances.
        """
        return getattr(self_or_cls, "__parameters__", tuple())

    @property
    def fields(self) -> dict[str, Field]:
        """Fields getter"""
        return self._fields

    def __setitem__(self, key: str, value: Any) -> None:
        try:
            self._fields[key].value = value
        except KeyError:
            raise KeyError(f"{type(self).__name__!r} does not contain the {key!r} field")

    def __getitem__(self, item: str) -> Any:
        try:
            return self._fields[item].value
        except KeyError:
            raise KeyError(f"{type(self).__name__!r} does not contain the {item!r} field.")

    def __setattr__(self, key: str, value: Any) -> None:
        if key.startswith("_"):
            object.__setattr__(self, key, value)
            return

        if key not in self._fields:
            raise AttributeError(f"{type(self).__name__!r} does not contain the {key!r} attribute.")

        self[key] = value

    def __getattr__(self, item: str) -> Any:
        if item.startswith("_") or item not in self._fields:
            raise AttributeError(f"{type(self).__name__!r} does not contain the {item!r} attribute.")

        return self[item]

    # noinspection PyMethodParameters
    @property_or_classproperty
    def avro_schema(self_or_cls) -> list[dict[str, Any]]:
        """Compute the avro schema of the model.

        :return: A dictionary object.
        """
        # noinspection PyTypeChecker
        encoder = AvroSchemaEncoder()
        return encoder.build(self_or_cls)

    @property
    def avro_data(self) -> dict[str, Any]:
        """Compute the avro data of the model.

        :return: A dictionary object.
        """
        encoder = AvroDataEncoder()
        return encoder.build(self)

    @property
    def avro_str(self) -> str:
        """Generate bytes representation of the current instance.

        :return: A bytes object.
        """
        # noinspection PyTypeChecker
        return b64encode(self.avro_bytes).decode()

    @property
    def avro_bytes(self) -> bytes:
        """Generate bytes representation of the current instance.

        :return: A bytes object.
        """
        # noinspection PyTypeChecker
        return MinosAvroProtocol.encode(self.avro_data, self.avro_schema)

    # noinspection PyUnusedLocal
    @staticmethod
    def encode_schema(encoder: SchemaEncoder, target: Any, **kwargs) -> Any:
        """Encode schema with the given encoder.

        :param encoder: The encoder instance.
        :param target: An optional pre-encoded schema.
        :param kwargs: Additional named arguments.
        :return: The encoded schema of the instance.
        """
        return MissingSentinel

    # noinspection PyUnusedLocal
    @staticmethod
    def decode_schema(decoder: SchemaDecoder, target: Any, **kwargs) -> Any:
        """Decode schema with the given encoder.

        :param decoder: The decoder instance.
        :param target: The schema to be decoded.
        :param kwargs: Additional named arguments.
        :return: The decoded schema as a type.
        """
        return MissingSentinel

    # noinspection PyUnusedLocal
    @staticmethod
    def encode_data(encoder: DataEncoder, target: Any, **kwargs) -> Any:
        """Encode data with the given encoder.

        :param encoder: The encoder instance.
        :param target: An optional pre-encoded data.
        :param kwargs: Additional named arguments.
        :return: The encoded data of the instance.
        """
        return MissingSentinel

    # noinspection PyUnusedLocal
    @staticmethod
    def decode_data(decoder: DataDecoder, target: Any, type_: ModelType, **kwargs) -> Any:
        """Decode data with the given decoder.

        :param decoder: The decoder instance.
        :param target: The data to be decoded.
        :param type_: The data type.
        :param kwargs: Additional named arguments.
        :return: A decoded instance.
        """
        return MissingSentinel

    def __eq__(self: T, other: T) -> bool:
        if type(self) == type(other) and self.fields == other.fields:
            return True

        if not self.__eq_reversing:
            try:
                self.__eq_reversing = True
                return other == self
            finally:
                self.__eq_reversing = False

        return False

    def __hash__(self) -> int:
        return hash(tuple(self.fields.values()))

    def __iter__(self) -> Iterable[str]:
        yield from self.fields.keys()

    def __len__(self):
        return len(self.fields)

    def __repr__(self) -> str:
        fields_repr = ", ".join(repr(field) for field in self.fields.values())
        return f"{type(self).__name__}({fields_repr})"


T = TypeVar("T", bound=Model)



from __future__ import (
    annotations,
)

from abc import (
    ABC,
)
from collections.abc import (
    Hashable,
)
from contextlib import (
    AbstractAsyncContextManager,
)

from cached_property import (
    cached_property,
)

from .injections import (
    Injectable,
)
from .pools import (
    Pool,
)


class Lock(AbstractAsyncContextManager):
    """Lock base class."""

    key: Hashable

    def __init__(self, key: Hashable, *args, **kwargs):
        if not isinstance(key, Hashable):
            raise ValueError(f"The key must be hashable. Obtained: {key!r} ({type(key)})")

        self.key = key

    @cached_property
    def hashed_key(self) -> int:
        """Get the hashed key.

        :return: An integer value.
        """
        if not isinstance(self.key, int):
            return hash(self.key)
        return self.key


@Injectable("lock_pool")
class LockPool(Pool[Lock], ABC):
    """Postgres Locking Pool class."""



from __future__ import (
    annotations,
)

import logging
from collections.abc import (
    Iterable,
)
from typing import (
    Any,
    Optional,
    Union,
    get_args,
    get_origin,
)

from .comparators import (
    TypeHintComparator,
    is_model_type,
)
from .model_types import (
    ModelType,
)

logger = logging.getLogger(__name__)


def build_union(options: Iterable[type, ...]) -> type:
    """Build the union type base on the given options.

    :param options: A tuple of types.
    :return: The union of types.
    """

    return Union[tuple(options)]


class TypeHintBuilder:
    """Type Hint Builder class."""

    def __init__(self, value: Any, type_: Optional[type] = None):
        self.value = value
        self.type_ = type_

    def build(self) -> type:
        """Build type hint from an instance..

        :return: A type.
        """
        return self._build(self.value, self.type_)

    def _build(self, value, type_: Optional[type]) -> type:

        if type_ is not None:
            if get_origin(type_) is Union:
                dynamic = self._build(value, None)
                options = tuple(self._build_from_dynamic(dynamic, static) for static in get_args(type_))
                return build_union(options)

        if isinstance(value, (tuple, list, set)):
            b1 = Any if (type_ is None or len(get_args(type_)) != 1) else get_args(type_)[0]
            return type(value)[self._build_from_iterable(value, b1)]

        if isinstance(value, dict):
            b1, b2 = (str, Any) if (type_ is None or len(get_args(type_)) != 2) else get_args(type_)
            return type(value)[
                self._build_from_iterable(value.keys(), b1), self._build_from_iterable(value.values(), b2)
            ]

        if is_model_type(value):
            return ModelType.from_model(value)

        if type_ is not None:
            dynamic = self._build(value, None)
            return self._build_from_dynamic(dynamic, type_)

        return type(value)

    def _build_from_iterable(self, values: Iterable, type_: Optional[type]) -> type:
        values = tuple(values)
        if len(values) == 0:
            return type_

        options = tuple(self._build(value, type_) for value in values)
        return build_union(options)

    @staticmethod
    def _build_from_dynamic(dynamic: type, static: Optional[type]) -> type:
        return dynamic if not len(get_args(static)) and TypeHintComparator(dynamic, static).match() else static



import io
from typing import (
    Any,
    Union,
)

from fastavro import (
    parse_schema,
    reader,
    writer,
)

from ...exceptions import (
    MinosProtocolException,
)
from ..abc import (
    MinosBinaryProtocol,
)


class MinosAvroProtocol(MinosBinaryProtocol):
    """Minos Avro Protocol class."""

    @classmethod
    def encode(cls, value: Any, schema: Any, *args, batch_mode: bool = False, **kwargs) -> bytes:
        """Encoder in avro for database Values
        all the headers are converted in fields with double underscore name
        the body is a set fields coming from the data type.

        :param value: The data to be stored.
        :param schema: The schema relative to the data.
        :param args: Additional positional arguments.
        :param batch_mode: If ``True`` the data is processed as a list of models, otherwise the data is processed as a
            single model.
        :param kwargs: Additional named arguments.
        :return: A bytes object.
        """
        if not batch_mode:
            value = [value]

        if not isinstance(schema, list):
            schema = [schema]

        try:
            raw_schema = cls._parse_schema(schema)
            return cls._write_data(value, raw_schema)
        except Exception as exc:
            raise MinosProtocolException(f"Error encoding data: {exc!r}")

    @staticmethod
    def _parse_schema(schema: list[dict[str, Any]]) -> dict[str, Any]:
        named_schemas = {}
        for item in schema[1::-1]:
            parse_schema(item, named_schemas)
        return parse_schema(schema[0], named_schemas, expand=True)

    @staticmethod
    def _write_data(value: list[dict[str, Any]], schema: dict[str, Any]):
        with io.BytesIO() as file:
            writer(file, schema, value)
            file.seek(0)
            content = file.getvalue()
        return content

    @classmethod
    def decode(cls, data: bytes, *args, batch_mode: bool = False, **kwargs) -> Any:
        """Decode the given bytes of data into a single dictionary or a sequence of dictionaries.

        :param data: A bytes object.
        :param args: Additional positional arguments.
        :param batch_mode: If ``True`` the data is processed as a list of models, otherwise the data is processed as a
            single model.
        :param kwargs: Additional named arguments.
        :return: A dictionary or a list of dictionaries.
        """

        try:
            with io.BytesIO(data) as file:
                ans = list(reader(file))
        except Exception as exc:
            raise MinosProtocolException(f"Error decoding the avro bytes: {exc}")

        if not batch_mode:
            if len(ans) > 1:
                raise MinosProtocolException(
                    f"The 'batch_mode' argument was set to {False!r} but data contains multiple values: {ans!r}"
                )
            ans = ans[0]

        return ans

    # noinspection PyUnusedLocal
    @classmethod
    def decode_schema(cls, data: bytes, *args, **kwargs) -> Union[dict[str, Any], list[dict[str, Any]]]:
        """Decode the given bytes of data into a single dictionary or a sequence of dictionaries.

        :param data: A bytes object.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: A tuple or a list of tuples.
        """

        try:
            with io.BytesIO(data) as file:
                r = reader(file)
                schema = r.writer_schema

        except Exception as exc:
            raise MinosProtocolException(f"Error getting avro schema: {exc}")

        return schema



import logging
from abc import (
    ABC,
    abstractmethod,
)

from aiomisc import (
    Service,
)

logger = logging.getLogger(__name__)


class Port(Service, ABC):
    """Port base class."""

    async def start(self) -> None:
        """Start the port execution.

        :return: This method does not return anything.
        """
        try:
            return await self._start()
        except Exception as exc:
            logger.exception(f"Raised an exception on {self!r} start: {exc!r}")
            raise exc

    @abstractmethod
    def _start(self):
        raise NotImplementedError

    async def stop(self, err: Exception = None) -> None:
        """Stop the port execution.

        :param err: Optional exception that stopped the execution.
        :return: This method does not return anything.
        """
        try:
            return await self._stop(err)
        except Exception as exc:
            logger.exception(f"Raised an exception on {self!r} stop: {exc!r}")
            raise exc

    @abstractmethod
    async def _stop(self, err: Exception = None) -> None:
        raise NotImplementedError



from __future__ import (
    annotations,
)

from typing import (
    Any,
    Type,
)


class MinosException(Exception):
    """Exception class for import packages or modules"""

    __slots__ = "_message"

    def __init__(self, error_message: str):
        self._message = error_message

    def __repr__(self):
        return f"{type(self).__name__}(message={repr(self._message)})"

    def __str__(self) -> str:
        """represent in a string format the error message passed during the instantiation"""
        return self._message


class NotProvidedException(MinosException):
    """Exception to be raised when a dependency is needed but not provided."""


class MinosImportException(MinosException):
    pass


class MinosProtocolException(MinosException):
    pass


class MinosMessageException(MinosException):
    pass


class MinosConfigException(MinosException):
    """Base config exception."""


class MinosBrokerException(MinosException):
    """Base broker exception"""


class MinosHandlerException(MinosException):
    """Base handler exception"""


class MinosLockException(MinosException):
    """Base lock exception"""


class MinosModelException(MinosException):
    """Exception to be raised when some mandatory condition is not satisfied by a model."""

    pass


class EmptyMinosModelSequenceException(MinosModelException):
    """Exception to be raised when a sequence must be not empty, but it is empty."""

    pass


class MultiTypeMinosModelSequenceException(MinosModelException):
    """Exception to be raised when a sequence doesn't satisfy the condition to have the same type for each item."""

    pass


class MinosModelAttributeException(MinosException):
    """Base model attributes exception."""

    pass


class MinosReqAttributeException(MinosModelAttributeException):
    """Exception to be raised when some required attributes are not provided."""

    pass


class MinosTypeAttributeException(MinosModelAttributeException):
    """Exception to be raised when there are any mismatching between the expected and observed attribute type."""

    def __init__(self, name: str, target_type: Type, value: Any):
        self.name = name
        self.target_type = target_type
        self.value = value
        super().__init__(
            f"The {target_type!r} expected type for {name!r} does not match with "
            f"the given data type: {type(value)!r} ({value!r})"
        )


class MinosMalformedAttributeException(MinosModelAttributeException):
    """Exception to be raised when there are any kind of problems with the type definition."""

    pass


class MinosParseAttributeException(MinosModelAttributeException):
    """Exception to be raised when there are any kind of problems with the parsing logic."""

    def __init__(self, name: str, value: Any, exception: Exception):
        self.name = name
        self.value = value
        self.exception = exception
        super().__init__(f"{repr(exception)} was raised while parsing {repr(name)} field with {repr(value)} value.")


class MinosAttributeValidationException(MinosModelAttributeException):
    """Exception to be raised when some fields are not valid."""

    def __init__(self, name: str, value: Any):
        self.name = name
        self.value = value
        super().__init__(f"{repr(value)} value does not pass the {repr(name)} field validation.")


class DataDecoderException(MinosModelException):
    """Base data decoder exception."""


class DataDecoderMalformedTypeException(DataDecoderException):
    """Exception to be raised when malformed types are provided."""


class DataDecoderRequiredValueException(DataDecoderException):
    """Exception to be raised when required values are not provided."""


class DataDecoderTypeException(DataDecoderException):
    """Exception to be raised when expected and provided types do not match."""

    def __init__(self, target_type: Type, value: Any):
        self.target_type = target_type
        self.value = value
        super().__init__(
            f"The {target_type!r} expected type does not match the given data type: {type(value)!r} ({value!r})"
        )



from __future__ import (
    annotations,
)

import logging
from itertools import (
    zip_longest,
)
from typing import (
    Iterable,
    Type,
    TypeVar,
    Union,
)

from ..abc import (
    Model,
)
from ..fields import (
    Field,
)
from ..types import (
    MissingSentinel,
    ModelType,
)

logger = logging.getLogger(__name__)


class DynamicModel(Model):
    """Base class for ``minos`` dynamic model entities"""

    def __init__(self, fields: Union[Iterable[Field], dict[str, Field]], **kwargs):
        super().__init__(fields)

    @classmethod
    def from_model_type(cls: Type[T], model_type: ModelType, *args, **kwargs) -> T:
        """Build a ``DynamicModel`` from a ``ModelType``.

        :param model_type: ``ModelType`` object containing the model structure
        :param args: Positional arguments to be passed to the model constructor.
        :param kwargs: Named arguments to be passed to the model constructor.
        :return: A new ``DynamicModel`` instance.
        """
        fields = cls._build_fields(model_type.type_hints, *args, **kwargs)
        return cls(fields=fields)

    @classmethod
    def _build_fields(cls, type_hints: dict[str, type], *args, **kwargs) -> dict[str, Field]:
        fields = dict()
        for (name, type_val), value in zip_longest(type_hints.items(), args, fillvalue=MissingSentinel):
            if name in kwargs and value is not MissingSentinel:
                raise TypeError(f"got multiple values for argument {repr(name)}")

            if value is MissingSentinel and name in kwargs:
                value = kwargs[name]

            fields[name] = cls._field_cls(name, type_val, value)
        return fields


T = TypeVar("T", bound=DynamicModel)



from collections.abc import (
    Hashable,
)
from typing import (
    Any,
    AsyncContextManager,
    AsyncIterator,
    Optional,
)

from aiomisc.pool import (
    ContextManager,
)
from aiopg import (
    Cursor,
)

from ..injections import (
    Inject,
)
from ..setup import (
    SetupMixin,
)
from .locks import (
    PostgreSqlLock,
)
from .pools import (
    PostgreSqlPool,
)


class PostgreSqlMinosDatabase(SetupMixin):
    """PostgreSql Minos Database base class."""

    def __init__(
        self,
        database: str,
        host: Optional[str] = None,
        port: Optional[int] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._database = database
        self._host = host
        self._port = port
        self._user = user
        self._password = password

        self._pool = None
        self._owned_pool = False

    @property
    def database(self) -> str:
        """Get the database's database.

        :return: A ``str`` value.
        """
        return self.pool.database

    @property
    def host(self) -> str:
        """Get the database's host.

        :return: A ``str`` value.
        """
        return self.pool.host

    @property
    def port(self) -> int:
        """Get the database's port.

        :return: An ``int`` value.
        """
        return self.pool.port

    @property
    def user(self) -> str:
        """Get the database's user.

        :return: A ``str`` value.
        """
        return self.pool.user

    @property
    def password(self) -> str:
        """Get the database's password.

        :return: A ``str`` value.
        """
        return self.pool.password

    async def _destroy(self) -> None:
        if self._owned_pool:
            await self._pool.destroy()
            self._pool = None
            self._owned_pool = False

    async def submit_query_and_fetchone(self, *args, **kwargs) -> tuple:
        """Submit a SQL query and gets the first response.

        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """
        return await self.submit_query_and_iter(*args, **kwargs).__anext__()

    # noinspection PyUnusedLocal
    async def submit_query_and_iter(
        self,
        operation: Any,
        parameters: Any = None,
        *,
        timeout: Optional[float] = None,
        lock: Optional[int] = None,
        streaming_mode: bool = False,
        **kwargs,
    ) -> AsyncIterator[tuple]:
        """Submit a SQL query and return an asynchronous iterator.

        :param operation: Query to be executed.
        :param parameters: Parameters to be projected into the query.
        :param timeout: An optional timeout.
        :param lock: Optional key to perform the query with locking. If not set, the query is performed without any
            lock.
        :param streaming_mode: If ``True`` the data fetching is performed in streaming mode, that is iterating over the
            cursor and yielding once a time (requires an opening connection to do that). Otherwise, all the data is
            fetched and keep in memory before yielding it.
        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """
        if lock is None:
            context_manager = self.cursor()
        else:
            context_manager = self.locked_cursor(lock)

        async with context_manager as cursor:
            await cursor.execute(operation=operation, parameters=parameters, timeout=timeout)

            if streaming_mode:
                async for row in cursor:
                    yield row
                return

            rows = await cursor.fetchall()

        for row in rows:
            yield row

    # noinspection PyUnusedLocal
    async def submit_query(
        self, operation: Any, parameters: Any = None, *, timeout: Optional[float] = None, lock: Any = None, **kwargs
    ) -> None:
        """Submit a SQL query.

        :param operation: Query to be executed.
        :param parameters: Parameters to be projected into the query.
        :param timeout: An optional timeout.
        :param lock: Optional key to perform the query with locking. If not set, the query is performed without any
            lock.
        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """
        if lock is None:
            context_manager = self.cursor()
        else:
            context_manager = self.locked_cursor(lock)

        async with context_manager as cursor:
            await cursor.execute(operation=operation, parameters=parameters, timeout=timeout)

    def locked_cursor(self, key: Hashable, *args, **kwargs) -> AsyncContextManager[Cursor]:
        """Get a new locked cursor.

        :param key: The key to be used for locking.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: A Cursor wrapped into an asynchronous context manager.
        """
        lock = PostgreSqlLock(self.pool.acquire(), key, *args, **kwargs)

        async def _fn_enter():
            await lock.__aenter__()
            return lock.cursor

        async def _fn_exit(_):
            await lock.__aexit__(None, None, None)

        return ContextManager(_fn_enter, _fn_exit)

    def cursor(self, *args, **kwargs) -> AsyncContextManager[Cursor]:
        """Get a new cursor.

        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: A Cursor wrapped into an asynchronous context manager.
        """
        acquired = self.pool.acquire()

        async def _fn_enter():
            connection = await acquired.__aenter__()
            cursor = await connection.cursor(*args, **kwargs).__aenter__()
            return cursor

        async def _fn_exit(cursor: Cursor):
            if not cursor.closed:
                cursor.close()
            await acquired.__aexit__(None, None, None)

        return ContextManager(_fn_enter, _fn_exit)

    @property
    def pool(self) -> PostgreSqlPool:
        """Get the connections pool.

        :return: A ``Pool`` object.
        """
        if self._pool is None:
            self._pool, self._owned_pool = self._build_pool()
        return self._pool

    @Inject()
    def _build_pool(self, pool: PostgreSqlPool = None) -> tuple[PostgreSqlPool, bool]:
        if pool is not None:
            return pool, False

        pool = PostgreSqlPool(
            host=self._host, port=self._port, database=self._database, user=self._user, password=self._password
        )
        return pool, True



"""The common core of the Minos Framework."""
__author__ = "Minos Framework Devs"
__email__ = "hey@minos.run"
__version__ = "0.7.0.dev1"

from .builders import (
    BuildableMixin,
    Builder,
)
from .config import (
    Config,
    ConfigV1,
    ConfigV2,
    MinosConfig,
)
from .database import (
    PostgreSqlLock,
    PostgreSqlLockPool,
    PostgreSqlMinosDatabase,
    PostgreSqlPool,
)
from .datetime import (
    NULL_DATETIME,
    current_datetime,
)
from .exceptions import (
    DataDecoderException,
    DataDecoderMalformedTypeException,
    DataDecoderRequiredValueException,
    DataDecoderTypeException,
    EmptyMinosModelSequenceException,
    MinosAttributeValidationException,
    MinosBrokerException,
    MinosConfigException,
    MinosException,
    MinosHandlerException,
    MinosImportException,
    MinosLockException,
    MinosMalformedAttributeException,
    MinosMessageException,
    MinosModelAttributeException,
    MinosModelException,
    MinosParseAttributeException,
    MinosProtocolException,
    MinosReqAttributeException,
    MinosTypeAttributeException,
    MultiTypeMinosModelSequenceException,
    NotProvidedException,
)
from .importlib import (
    classname,
    get_internal_modules,
    import_module,
)
from .injections import (
    DependencyInjector,
    Inject,
    Injectable,
    InjectableMixin,
)
from .launchers import (
    EntrypointLauncher,
)
from .locks import (
    Lock,
    LockPool,
)
from .meta import (
    classproperty,
    property_or_classproperty,
    self_or_classmethod,
)
from .model import (
    AvroDataDecoder,
    AvroDataEncoder,
    AvroSchemaDecoder,
    AvroSchemaEncoder,
    BucketModel,
    DataDecoder,
    DataEncoder,
    DataTransferObject,
    DeclarativeModel,
    DynamicModel,
    Field,
    GenericTypeProjector,
    MinosModel,
    MissingSentinel,
    Model,
    ModelField,
    ModelType,
    NoneType,
    SchemaDecoder,
    SchemaEncoder,
    TypeHintBuilder,
    TypeHintComparator,
    is_model_type,
)
from .object import (
    Object,
)
from .pools import (
    MinosPool,
    Pool,
)
from .ports import (
    Port,
)
from .protocol import (
    MinosAvroDatabaseProtocol,
    MinosAvroMessageProtocol,
    MinosAvroProtocol,
    MinosBinaryProtocol,
    MinosJsonBinaryProtocol,
)
from .retries import (
    CircuitBreakerMixin,
)
from .setup import (
    MinosSetup,
    SetupMixin,
)
from .storage import (
    MinosStorage,
    MinosStorageLmdb,
)
from .uuid import (
    NULL_UUID,
    UUID_REGEX,
)



from __future__ import (
    annotations,
)

import logging
from datetime import (
    date,
    datetime,
    time,
    timedelta,
    timezone,
)
from decimal import (
    Decimal,
)
from typing import (
    TYPE_CHECKING,
    Any,
)
from uuid import (
    UUID,
)

from .....exceptions import (
    MinosMalformedAttributeException,
)
from ....types import (
    MissingSentinel,
)
from ...abc import (
    DataEncoder,
)

if TYPE_CHECKING:
    from ....abc import (
        Model,
    )
    from ....fields import (
        Field,
    )

logger = logging.getLogger(__name__)


class AvroDataEncoder(DataEncoder):
    """Avro Data Encoder class."""

    def __init__(self, value: Any = None):
        self.value = value

    def build(self, value=MissingSentinel, **kwargs) -> Any:
        """Build an avro data representation based on the content of the given field.

        :param value: The value to be encoded.
        :return: A `avro`-compatible data.
        """
        if value is MissingSentinel:
            value = self.value
        return self._build(value, **kwargs)

    def _build(self, value: Any, **kwargs) -> Any:
        if value is None:
            return None

        from ....abc import (
            Model,
        )

        if isinstance(value, Model):
            return self._build_model(value, **kwargs)

        from ....abc import (
            Field,
        )

        if isinstance(value, Field):
            return self._build_field(value, **kwargs)

        if isinstance(value, (str, int, bool, float, bytes)):
            return value

        if isinstance(value, memoryview):
            return value.tobytes()

        if isinstance(value, Decimal):
            return float(value)

        if isinstance(value, datetime):
            return self._build_datetime(value, **kwargs)

        if isinstance(value, timedelta):
            return self._build_timedelta(value, **kwargs)

        if isinstance(value, date):
            return self._build_date(value, **kwargs)

        if isinstance(value, time):
            return self._build_time(value, **kwargs)

        if isinstance(value, UUID):
            return self._build_uuid(value, **kwargs)

        if isinstance(value, (list, set)):
            return [self._build(v, **kwargs) for v in value]

        if isinstance(value, dict):
            return {k: self._build(v, **kwargs) for k, v in value.items()}

        raise MinosMalformedAttributeException(f"Given type is not supported: {type(value)!r} ({value!r})")

    def _build_model(self, model: Model, **kwargs) -> Any:
        raw = {name: self._build_field(field, **kwargs) for name, field in model.fields.items()}

        if (ans := model.encode_data(self, raw, **kwargs)) is not MissingSentinel:
            return ans

        return self._build(raw, **kwargs)

    def _build_field(self, field: Field, **kwargs) -> Any:
        return self._build(field.value, **kwargs)

    @staticmethod
    def _build_date(value: date, **kwargs) -> int:
        return (value - date(1970, 1, 1)).days

    @staticmethod
    def _build_time(value: time, **kwargs) -> int:
        return (datetime.combine(date(1, 1, 1), value) - datetime(1, 1, 1)) // timedelta(microseconds=1)

    @staticmethod
    def _build_datetime(value: datetime, **kwargs) -> int:
        return (value.astimezone(timezone.utc) - datetime(1970, 1, 1, tzinfo=timezone.utc)) // timedelta(microseconds=1)

    @staticmethod
    def _build_timedelta(value: timedelta, **kwargs) -> int:
        return value // timedelta(microseconds=1)

    @staticmethod
    def _build_uuid(value: UUID, **kwargs) -> str:
        return str(value)



from __future__ import (
    annotations,
)

from functools import (
    lru_cache,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    NamedTuple,
    Optional,
    Type,
    Union,
)

from ...exceptions import (
    MinosImportException,
)
from ...importlib import (
    import_module,
)
from .generics import (
    GenericTypeProjector,
)

if TYPE_CHECKING:
    from ..abc import (
        Model,
    )


class ModelType(type):
    """Model Type class."""

    name: str
    namespace: str
    type_hints: dict[str, Type]

    @classmethod
    def build(
        mcs, name_: str, type_hints_: Optional[dict[str, type]] = None, *, namespace_: Optional[str] = None, **kwargs
    ) -> ModelType:
        """Build a new ``ModelType`` instance.

        :param name_: Name of the new type.
        :param type_hints_: Type hints of the new type.
        :param namespace_: Namespace of the new type.
        :param kwargs: Type hints of the new type as named parameters.
        :return: A ``ModelType`` instance.
        """
        if type_hints_ is None:
            type_hints_ = tuple(kwargs.items())
        else:
            if len(kwargs):
                raise ValueError("Type hints can be passed in a dictionary or as named parameters, but not both.")
            type_hints_ = tuple(type_hints_.items())

        if namespace_ is None:
            try:
                namespace_, name_ = name_.rsplit(".", 1)
            except ValueError:
                namespace_ = str()

        # noinspection PyTypeChecker
        return mcs._build(name_, type_hints_, namespace_)

    @classmethod
    @lru_cache()
    def _build(mcs, name_: str, type_hints_: tuple[tuple[str, type], ...], namespace_: Optional[str]):
        return mcs(name_, tuple(), {"type_hints": dict(type_hints_), "namespace": namespace_})

    @classmethod
    def from_typed_dict(mcs, typed_dict) -> ModelType:
        """Build a new ``ModelType`` instance from a ``typing.TypedDict``.

        :param typed_dict: Typed dict to be used as base.
        :return: A ``ModelType`` instance.
        """
        return mcs.build(typed_dict.__name__, typed_dict.__annotations__)

    @staticmethod
    def from_model(model: Union[Model, type[Model]]) -> ModelType:
        """Build a new instance from model class.

        :param model: The model class.
        :return: A new ``ModelType`` instance.
        """
        # noinspection PyTypeChecker
        return ModelType.build(name_=model.classname, type_hints_=GenericTypeProjector.from_model(model).build())

    def __call__(cls, *args, **kwargs) -> Model:
        return cls.model_cls.from_model_type(cls, *args, **kwargs)

    @property
    def model_cls(cls) -> Type[Model]:
        """Get the model class if defined or ``DataTransferObject`` otherwise.

        :return: A model class.
        """
        try:
            # noinspection PyTypeChecker
            return import_module(cls.classname)
        except MinosImportException:
            from ..dynamic import (
                DataTransferObject,
            )

            return DataTransferObject

    @property
    def name(cls) -> str:
        """Get the type name.

        :return: A string object.
        """
        return cls.__name__

    @property
    def classname(cls) -> str:
        """Get the full class name.

        :return: An string object.
        """
        if not len(cls.namespace):
            return cls.name
        return f"{cls.namespace}.{cls.name}"

    def __le__(cls, other: Any) -> bool:
        from .comparators import (
            TypeHintComparator,
        )

        return type(cls).__eq__(cls, other) or (
            type(cls) == type(other)
            and cls.name == other.name
            and cls.namespace == other.namespace
            and set(cls.type_hints.keys()) <= set(other.type_hints.keys())
            and all(TypeHintComparator(v, other.type_hints[k]).match() for k, v in cls.type_hints.items())
        )

    def __lt__(cls, other: Any) -> bool:
        from .comparators import (
            TypeHintComparator,
        )

        return (
            type(cls) == type(other)
            and cls.name == other.name
            and cls.namespace == other.namespace
            and set(cls.type_hints.keys()) < set(other.type_hints.keys())
            and all(TypeHintComparator(v, other.type_hints[k]).match() for k, v in cls.type_hints.items())
        )

    def __ge__(cls, other: Any) -> bool:
        from .comparators import (
            TypeHintComparator,
        )

        return type(cls).__eq__(cls, other) or (
            type(cls) == type(other)
            and cls.name == other.name
            and cls.namespace == other.namespace
            and set(cls.type_hints.keys()) >= set(other.type_hints.keys())
            and all(TypeHintComparator(v, cls.type_hints[k]).match() for k, v in other.type_hints.items())
        )

    def __gt__(cls, other: Any) -> bool:
        from .comparators import (
            TypeHintComparator,
        )

        return (
            type(cls) == type(other)
            and cls.name == other.name
            and cls.namespace == other.namespace
            and set(cls.type_hints.keys()) > set(other.type_hints.keys())
            and all(TypeHintComparator(v, cls.type_hints[k]).match() for k, v in other.type_hints.items())
        )

    def __eq__(cls, other: Any) -> bool:
        conditions = (
            cls._equal_with_model_type,
            cls._equal_with_model,
            cls._equal_with_inherited_model,
            cls._equal_with_bucket_model,
        )
        # noinspection PyArgumentList
        return any(condition(other) for condition in conditions)

    def _equal_with_model_type(cls, other: ModelType) -> bool:
        from .comparators import (
            TypeHintComparator,
        )

        return (
            type(cls) == type(other)
            and cls.name == other.name
            and cls.namespace == other.namespace
            and set(cls.type_hints.keys()) == set(other.type_hints.keys())
            and all(TypeHintComparator(v, other.type_hints[k]).match() for k, v in cls.type_hints.items())
        )

    def _equal_with_model(cls, other: Any) -> bool:
        return hasattr(other, "model_type") and cls == ModelType.from_model(other)

    def _equal_with_inherited_model(cls, other: ModelType) -> bool:
        return (
            type(cls) == type(other) and cls.model_cls != other.model_cls and issubclass(cls.model_cls, other.model_cls)
        )

    def _equal_with_bucket_model(self, other: Any) -> bool:
        from ..dynamic import (
            BucketModel,
        )

        return (
            hasattr(other, "model_cls")
            and issubclass(self.model_cls, other.model_cls)
            and issubclass(other.model_cls, BucketModel)
        )

    def __hash__(cls) -> int:
        return hash(tuple(cls))

    def __iter__(cls) -> Iterable:
        # noinspection PyRedundantParentheses
        yield from (cls.name, cls.namespace, tuple(cls.type_hints.items()))

    def __repr__(cls):
        return f"{type(cls).__name__}(name={cls.name!r}, namespace={cls.namespace!r}, type_hints={cls.type_hints!r})"


class FieldType(NamedTuple):
    """Field Type class."""

    name: str
    type: type



from __future__ import (
    annotations,
)

import logging
from collections.abc import (
    Iterable,
    Mapping,
)
from contextlib import (
    suppress,
)
from datetime import (
    date,
    datetime,
    time,
    timedelta,
    timezone,
)
from itertools import (
    zip_longest,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
)
from uuid import (
    UUID,
)

from .....exceptions import (
    DataDecoderMalformedTypeException,
    DataDecoderRequiredValueException,
    DataDecoderTypeException,
)
from ....types import (
    MissingSentinel,
    ModelType,
    NoneType,
    TypeHintBuilder,
    is_model_subclass,
    is_type_subclass,
    unpack_typevar,
)
from ...abc import (
    DataDecoder,
)

if TYPE_CHECKING:
    from ....abc import (
        Model,
    )

logger = logging.getLogger(__name__)


class AvroDataDecoder(DataDecoder):
    """Avro Data Decoder class."""

    def __init__(self, type_: Optional[type] = None):
        self.type_ = type_

    def build(self, data: Any, type_: Any = MissingSentinel, **kwargs) -> Any:
        """Cast data type according to the field definition.

        :param data: Data to be casted.
        :param type_: The type of the decoded data.
        :return: The casted object.
        """
        if type_ is MissingSentinel:
            type_ = self.type_
        return self._build(type_, data, **kwargs)

    def _build(self, type_: type, data: Any, **kwargs) -> Any:
        origin = get_origin(type_)
        if origin is not Union:
            return self._build_single(type_, data, **kwargs)
        return self._build_union(type_, data, **kwargs)

    def _build_union(self, type_: type, data: Any, **kwargs) -> Any:
        alternatives = get_args(type_)
        for alternative_type in alternatives:
            with suppress(Exception):
                return self._build_single(alternative_type, data, **kwargs)

        if type_ is not NoneType:
            if data is None:
                raise DataDecoderRequiredValueException(f"Value is {None!r}.")

            if data is MissingSentinel:
                raise DataDecoderRequiredValueException("Value is missing.")

        raise DataDecoderTypeException(type_, data)

    def _build_single(self, type_: type, data: Any, **kwargs) -> Any:
        if type_ is Any:
            type_ = TypeHintBuilder(data).build()
        if isinstance(type_, TypeVar):
            unpacked_type = unpack_typevar(type_)
            return self._build(unpacked_type, data, **kwargs)

        if type_ is NoneType:
            return self._build_none(type_, data, **kwargs)

        if data is None:
            raise DataDecoderRequiredValueException(f"Value is {None!r}.")

        if data is MissingSentinel:
            raise DataDecoderRequiredValueException("Value is missing.")

        if is_model_subclass(type_):
            # noinspection PyTypeChecker
            return self._build_model(type_, data, **kwargs)

        if is_type_subclass(type_):
            if issubclass(type_, bool):
                return self._build_bool(data, **kwargs)

            if issubclass(type_, int):
                return self._build_int(type_, data, **kwargs)

            if issubclass(type_, float):
                return self._build_float(data, **kwargs)

            if issubclass(type_, str):
                return self._build_string(type_, data, **kwargs)

            if issubclass(type_, bytes):
                return self._build_bytes(data, **kwargs)

            if issubclass(type_, datetime):
                return self._build_datetime(data, **kwargs)

            if issubclass(type_, timedelta):
                return self._build_timedelta(data, **kwargs)

            if issubclass(type_, date):
                return self._build_date(data, **kwargs)

            if issubclass(type_, time):
                return self._build_time(data, **kwargs)

            if issubclass(type_, UUID):
                return self._build_uuid(data, **kwargs)

            if isinstance(type_, ModelType):
                return self._build_model_type(type_, data, **kwargs)

        return self._build_collection(type_, data, **kwargs)

    @staticmethod
    def _build_none(type_: type, data: Any, **kwargs) -> Any:
        if data is None or data is MissingSentinel:
            return None

        raise DataDecoderTypeException(type_, data)

    @staticmethod
    def _build_int(type_, data: Any, **kwargs) -> int:
        try:
            return type_(data)
        except (ValueError, TypeError):
            raise DataDecoderTypeException(type_, data)

    @staticmethod
    def _build_float(data: Any, **kwargs) -> float:
        try:
            return float(data)
        except (ValueError, TypeError):
            raise DataDecoderTypeException(float, data)

    @staticmethod
    def _build_bool(data: Any, **kwargs) -> bool:
        if not isinstance(data, bool):
            raise DataDecoderTypeException(bool, data)
        return data

    @staticmethod
    def _build_string(type_, data: Any, **kwargs) -> str:
        if not isinstance(data, str):
            raise DataDecoderTypeException(str, data)
        return type_(data)

    @staticmethod
    def _build_bytes(data: Any, **kwargs) -> bytes:
        if not isinstance(data, bytes):
            raise DataDecoderTypeException(bytes, data)
        return data

    @staticmethod
    def _build_date(data: Any, **kwargs) -> date:
        if isinstance(data, date):
            return data
        elif isinstance(data, int):
            return date(1970, 1, 1) + timedelta(days=data)
        raise DataDecoderTypeException(date, data)

    @staticmethod
    def _build_time(data: Any, **kwargs) -> time:
        if isinstance(data, time):
            return data
        if isinstance(data, int):
            return (datetime(1, 1, 1) + timedelta(microseconds=data)).time()
        raise DataDecoderTypeException(time, data)

    @staticmethod
    def _build_datetime(data: Any, **kwargs) -> datetime:
        if isinstance(data, datetime):
            return data
        if isinstance(data, int):
            return datetime(1970, 1, 1, tzinfo=timezone.utc) + data * timedelta(microseconds=1)
        raise DataDecoderTypeException(datetime, data)

    @staticmethod
    def _build_timedelta(data: Any, **kwargs) -> timedelta:
        if isinstance(data, timedelta):
            return data
        if isinstance(data, int):
            return timedelta(microseconds=data)
        raise DataDecoderTypeException(timedelta, data)

    @staticmethod
    def _build_uuid(data: Any, **kwargs) -> UUID:
        if isinstance(data, UUID):
            return data
        elif isinstance(data, str):
            try:
                return UUID(hex=data)
            except ValueError:
                pass
        elif isinstance(data, bytes):
            try:
                return UUID(bytes=data)
            except ValueError:
                pass
        raise DataDecoderTypeException(UUID, data)

    def _build_model(self, type_: Type[Model], data: Any, **kwargs) -> Any:
        if is_type_subclass(type_) and isinstance(data, type_):
            return data
        return self._build_model_type(ModelType.from_model(type_), data, **kwargs)

    def _build_model_type(self, type_: ModelType, data: Any, **kwargs) -> Any:
        if hasattr(data, "model_type"):
            if ModelType.from_model(data) >= type_:
                return data

        if (ans := type_.model_cls.decode_data(self, data, type_, **kwargs)) is not MissingSentinel:
            return ans

        if isinstance(data, dict):
            with suppress(Exception):
                decoded_data = {
                    field_name: self._build(field_type, data[field_name], **kwargs)
                    for field_name, field_type in type_.type_hints.items()
                    if field_name in data
                }
                return type_(**decoded_data, additional_type_hints=type_.type_hints)

        with suppress(Exception):
            decoded_data = (
                self._build(field_type, field_value, **kwargs)
                for field_value, field_type in zip_longest((data,), type_.type_hints.values())
            )
            return type_(*decoded_data, additional_type_hints=type_.type_hints)

        raise DataDecoderTypeException(type_, data)

    def _build_collection(self, type_: type, data: Any, **kwargs) -> Any:
        origin_type = get_origin(type_)
        if origin_type is None:
            raise DataDecoderMalformedTypeException(f"Type is malformed. Obtained: '{type_}'.")

        if origin_type is list:
            return self._build_list(data, type_, **kwargs)

        if origin_type is set:
            return self._build_set(data, type_, **kwargs)

        if origin_type is dict:
            return self._build_dict(data, type_, **kwargs)

        raise DataDecoderTypeException(type_, data)

    def _build_list(self, data: Any, type_values: Any, **kwargs) -> list[Any]:
        type_values = get_args(type_values)[0]
        if isinstance(data, str) or not isinstance(data, Iterable):
            raise DataDecoderTypeException(list, data)

        return list(self._build_iterable(data, type_values, **kwargs))

    def _build_set(self, data: Any, type_values: Any, **kwargs) -> set[Any]:
        type_values = get_args(type_values)[0]
        if isinstance(data, str) or not isinstance(data, Iterable):
            raise DataDecoderTypeException(set, data)

        return set(self._build_iterable(data, type_values, **kwargs))

    def _build_dict(self, data: dict, type_: type, **kwargs) -> dict[str, Any]:
        type_keys, type_values = get_args(type_)
        if not isinstance(data, Mapping):
            raise DataDecoderTypeException(dict, data)

        if type_keys is not str:
            raise DataDecoderMalformedTypeException(f"dictionary keys must be {str!r}. Obtained: {type_keys!r}")

        keys = self._build_iterable(data.keys(), type_keys, **kwargs)
        values = self._build_iterable(data.values(), type_values, **kwargs)
        return dict(zip(keys, values))

    def _build_iterable(self, data: Iterable, type_params: type, **kwargs) -> Iterable:
        return (self._build(type_params, item, **kwargs) for item in data)



from __future__ import (
    annotations,
)

from typing import (
    Type,
    TypeVar,
)

from .abc import (
    DynamicModel,
)


class BucketModel(DynamicModel):
    """Bucket Model class."""

    @classmethod
    def empty(cls: Type[T]) -> T:
        """Build an empty ``BucketModel`` instance.

        :return: A ``BucketModel`` instance.
        """
        return cls(fields=dict())


T = TypeVar("T", bound=BucketModel)



from datetime import (
    datetime,
    timezone,
)


def current_datetime() -> datetime:
    """Get current datetime in `UTC`.

    :return: A ``datetime`` instance.
    """
    return datetime.now(tz=timezone.utc)


NULL_DATETIME = datetime.max.replace(tzinfo=timezone.utc)



from __future__ import (
    annotations,
)

from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

if TYPE_CHECKING:
    from ..abc import (
        Model,
    )


# noinspection SpellCheckingInspection
def unpack_typevar(value: TypeVar) -> type:
    """Unpack `TypeVar` into a union of possible types.

    :param value: A type var instance.
    :return: A union of types.
    """
    return Union[value.__constraints__ or (value.__bound__ or Any,)]


class GenericTypeProjector:
    """Generic Type Projector."""

    def __init__(self, type_hints: dict[str, type], mapper: dict[TypeVar, type]):
        self.type_hints = type_hints
        self.mapper = mapper

    @classmethod
    def from_model(cls, model: Union[Model, type[Model]]) -> GenericTypeProjector:
        """Build a new instance from model.

        :param model: The model class.
        :return: A ``GenericTypeProjector`` instance.
        """
        # noinspection PyTypeChecker
        generics_ = dict(zip(model.type_hints_parameters, get_args(model)))
        # noinspection PyTypeChecker
        return cls(model.type_hints, generics_)

    def build(self) -> dict[str, type]:
        """Builder a projection of type vars values.

        :return: A dict of type hints.
        """
        return {k: self._build(v) for k, v in self.type_hints.items()}

    def _build(self, type_: type) -> type:
        if isinstance(type_, TypeVar):
            return self.mapper.get(type_, unpack_typevar(type_))

        origin = get_origin(type_)
        if origin is None:
            return type_

        # noinspection PyUnresolvedReferences
        return self._build(origin)[tuple(self._build(arg) for arg in get_args(type_))]



from __future__ import (
    annotations,
)

import logging
from functools import (
    lru_cache,
)
from itertools import (
    zip_longest,
)
from typing import (
    Any,
    Iterator,
    Optional,
    TypeVar,
    get_type_hints,
)

from ..meta import (
    self_or_classmethod,
)
from .abc import (
    Model,
)
from .types import (
    MissingSentinel,
    ModelType,
    TypeHintComparator,
)

logger = logging.getLogger(__name__)


class DeclarativeModel(Model):
    """Base class for ``minos`` declarative model entities."""

    def __init__(self, *args, **kwargs):
        """Class constructor.

        :param kwargs: Named arguments to be set as model attributes.
        """
        super().__init__()
        self._build_fields(*args, **kwargs)

    # noinspection PyUnusedLocal
    @classmethod
    def from_model_type(cls: type[T], model_type: ModelType, *args, **kwargs) -> T:
        """Build a ``DeclarativeModel`` from a ``ModelType``.

        :param model_type: ``ModelType`` object containing the model structure.
        :param args: Positional arguments to be passed to the model constructor.
        :param kwargs: Named arguments to be passed to the model constructor.
        :return: A new ``DeclarativeModel`` instance.
        """
        return cls(*args, **kwargs)

    def _build_fields(self, *args, additional_type_hints: Optional[dict[str, type]] = None, **kwargs) -> None:
        for (name, type_val), value in zip_longest(
            self._type_hints(additional_type_hints), args, fillvalue=MissingSentinel
        ):
            if name in kwargs and value is not MissingSentinel:
                raise TypeError(f"got multiple values for argument {repr(name)}")

            if value is MissingSentinel and name in kwargs:
                value = kwargs[name]

            self._fields[name] = self._field_cls(
                name, type_val, value, getattr(self, f"parse_{name}", None), getattr(self, f"validate_{name}", None)
            )

    # noinspection PyMethodParameters
    @self_or_classmethod
    def _type_hints(self_or_cls, additional_type_hints: Optional[dict[str, type]] = None) -> Iterator[tuple[str, Any]]:
        type_hints = dict()
        if isinstance(self_or_cls, type):
            cls = self_or_cls
        else:
            cls = type(self_or_cls)
        for b in cls.__mro__[::-1]:
            list_fields = _get_class_type_hints(b)
            type_hints |= list_fields
        logger.debug(f"The obtained type hints are: {type_hints!r}")

        if additional_type_hints:
            for name, hint in additional_type_hints.items():
                if name not in type_hints or TypeHintComparator(hint, type_hints[name]).match():
                    type_hints[name] = hint

        type_hints |= super()._type_hints()
        yield from type_hints.items()


@lru_cache()
def _get_class_type_hints(b: type) -> dict[str, type]:
    return {k: v for k, v in get_type_hints(b).items() if not k.startswith("_")}


T = TypeVar("T", bound=DeclarativeModel)
MinosModel = DeclarativeModel



from __future__ import (
    annotations,
)

import inspect
import logging
from typing import (
    Any,
    Callable,
    Iterable,
    Optional,
)

from ..exceptions import (
    DataDecoderMalformedTypeException,
    DataDecoderRequiredValueException,
    DataDecoderTypeException,
    MinosAttributeValidationException,
    MinosMalformedAttributeException,
    MinosParseAttributeException,
    MinosReqAttributeException,
    MinosTypeAttributeException,
)
from .serializers import (
    AvroDataDecoder,
    AvroDataEncoder,
    AvroSchemaDecoder,
    AvroSchemaEncoder,
)
from .types import (
    MissingSentinel,
    TypeHintBuilder,
    TypeHintComparator,
)

logger = logging.getLogger(__name__)


class Field:
    """Represents a model field."""

    __slots__ = "_name", "_type", "_value", "_parser", "_validator"

    def __init__(
        self,
        name: str,
        type_: type,
        value: Any = MissingSentinel,
        parser: Optional[Callable[[Any], Any]] = None,
        validator: Optional[Callable[[Any], Any]] = None,
    ):
        self._name = name
        self._type = type_
        self._parser = parser
        self._validator = validator

        self.value = value

    @property
    def name(self) -> str:
        """Name getter."""
        return self._name

    @property
    def type(self) -> type:
        """Type getter."""
        return self._type

    @property
    def real_type(self) -> type:
        """Real Type getter."""
        return TypeHintBuilder(self.value, self.type).build()

    @property
    def parser(self) -> Optional[Callable[[Any], Any]]:
        """Parser getter."""
        return self._parser

    @property
    def _parser_function(self) -> Optional[Callable[[Any], Any]]:
        if self.parser is None:
            return None
        if inspect.ismethod(self.parser):
            # noinspection PyUnresolvedReferences
            return self.parser.__func__
        return self.parser

    @property
    def validator(self) -> Optional[Callable[[Any], Any]]:
        """Parser getter."""
        return self._validator

    @property
    def _validator_function(self) -> Optional[Callable[[Any], Any]]:
        if self.validator is None:
            return None
        if inspect.ismethod(self.validator):
            # noinspection PyUnresolvedReferences
            return self.validator.__func__

    @property
    def value(self) -> Any:
        """Value getter."""
        return self._value

    @value.setter
    def value(self, data: Any) -> None:
        """Check if the given value is correct and stores it if ``True``, otherwise raises an exception.

        :param data: new value.
        :return: This method does not return anything.
        """
        logger.debug(f"Setting {data!r} value to {self._name!r} field with {self._type!r} type...")

        if self._parser is not None:
            try:
                data = self.parser(data)
            except Exception as exc:
                raise MinosParseAttributeException(self.name, data, exc)

        try:
            value = AvroDataDecoder(self.type).build(data)
        except DataDecoderMalformedTypeException as exc:
            raise MinosMalformedAttributeException(f"{self.name!r} field is malformed. {exc}")
        except DataDecoderRequiredValueException as exc:
            raise MinosReqAttributeException(f"{self.name!r} field is required. {exc}")
        except DataDecoderTypeException:
            raise MinosTypeAttributeException(self.name, self.type, data)

        if self.validator is not None and value is not None and not self.validator(value):
            raise MinosAttributeValidationException(self.name, value)

        self._value = value

    @property
    def avro_schema(self) -> dict[str, Any]:
        """Compute the avro schema of the field.

        :return: A dictionary object.
        """
        encoder = AvroSchemaEncoder()
        return encoder.build(self)

    @property
    def avro_data(self) -> Any:
        """Compute the avro data of the model.

        :return: A dictionary object.
        """
        encoder = AvroDataEncoder()
        return encoder.build(self)

    @classmethod
    def from_avro(cls, schema: dict, value: Any) -> Field:
        """Build a ``Field`` instance from the avro information.

        :param schema: Field's schema.
        :param value: Field's value.
        :return: A ``Field`` instance.
        """
        type_val = AvroSchemaDecoder(schema).build()
        return cls(schema["name"], type_val, value)

    def __eq__(self, other: Field) -> bool:
        return (
            type(self) == type(other)
            and self.name == other.name
            and self.value == other.value
            and self._parser_function == other._parser_function
            and self._validator_function == other._validator_function
            and TypeHintComparator(self.type, other.type).match()
        )

    def __hash__(self) -> int:
        return hash(tuple(self))

    def __iter__(self) -> Iterable:
        # noinspection PyRedundantParentheses
        yield from (self.name, self.type, self.value, self._parser_function, self._validator_function)

    def __repr__(self) -> str:
        return f"{self.name}={self.value!r}"


ModelField = Field



from __future__ import (
    annotations,
)

import os
import warnings
from abc import (
    ABC,
    abstractmethod,
)
from collections import (
    defaultdict,
)
from collections.abc import (
    Callable,
)
from contextlib import (
    suppress,
)
from pathlib import (
    Path,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Union,
)

import yaml

from ..exceptions import (
    MinosConfigException,
)
from ..importlib import (
    import_module,
)
from ..injections import (
    Injectable,
)

if TYPE_CHECKING:
    from ..injections import (
        InjectableMixin,
    )

sentinel = object()


@Injectable("config")
class Config(ABC):
    """Config base class."""

    __slots__ = ("_file_path", "_data", "_with_environment", "_parameterized")

    DEFAULT_VALUES: dict[str, Any] = dict()

    def __init__(self, path: Union[Path, str], with_environment: bool = True, **kwargs):
        super().__init__()
        if isinstance(path, str):
            path = Path(path)

        self._file_path = path
        self._data = self._load(path)

        self._with_environment = with_environment
        self._parameterized = kwargs

    def __new__(cls, *args, **kwargs) -> Config:
        if cls not in (Config, MinosConfig):
            return super().__new__(cls)

        from .v1 import (
            ConfigV1,
        )
        from .v2 import (
            ConfigV2,
        )

        version_mapper = defaultdict(
            lambda: ConfigV1,
            {
                1: ConfigV1,
                2: ConfigV2,
            },
        )

        version = _get_version(*args, **kwargs)

        return super().__new__(version_mapper[version])

    @property
    def file_path(self) -> Path:
        """Get the config's file path.

        :return: A ``Path`` instance.
        """
        return self._file_path

    @staticmethod
    def _load(path: Path) -> dict[str, Any]:
        if not path.exists():
            raise MinosConfigException(f"Check if this path: {path} is correct")

        with path.open() as file:
            return yaml.load(file, Loader=yaml.FullLoader)

    @property
    def version(self) -> int:
        """Get the version value.

        :return: A ``int`` instance.
        """
        return self._version

    @property
    @abstractmethod
    def _version(self) -> int:
        raise NotImplementedError

    def get_name(self) -> str:
        """Get the name value.

        :return: A ``str`` instance.
        """
        return self._get_name()

    @abstractmethod
    def _get_name(self) -> str:
        raise NotImplementedError

    def get_injections(self) -> list[type[InjectableMixin]]:
        """Get the injections value.

        :return: A ``list`` of ``InjectableMixin`` types.
        """
        return self._get_injections()

    @abstractmethod
    def _get_injections(self) -> list[type[InjectableMixin]]:
        raise NotImplementedError

    def get_default_database(self):
        """Get the default database value.

        :return: A ``dict`` containing the database's config values.
        """
        return self.get_database_by_name("default")

    def get_database_by_name(self, name: str) -> dict[str, Any]:
        """Get the database value by name.

        :param name: The name of the database. If ``None`` is provided then the default database will be used.
        :return: A ``dict`` containing the database's config values.
        """
        databases = self.get_databases()

        if name not in databases:
            name = "default"

        return databases[name]

    def get_databases(self) -> dict[str, dict[str, Any]]:
        """Get all databases' values.

        :return: A mapping from database name to database's config values.
        """
        return self._get_databases()

    def _get_databases(self) -> dict[str, dict[str, Any]]:
        raise NotImplementedError

    def get_interface_by_name(self, name: str) -> dict[str, Any]:
        """Get the interface value by name.

        :param name: The name of the interface.
        :return: A ``dict`` containing the interface's config values.
        """

        interfaces = self.get_interfaces()

        if name not in interfaces:
            raise MinosConfigException(f"There is not a {name!r} interface.")

        return interfaces[name]

    def get_interfaces(self) -> dict[str, dict[str, Any]]:
        """Get all interfaces' values.

        :return: A mapping from interface name to interface's config values.
        """
        return self._get_interfaces()

    def _get_interfaces(self) -> dict[str, dict[str, Any]]:
        raise NotImplementedError

    def get_pools(self) -> dict[str, type]:
        """Get the pools value.

        :return: A ``dict`` with pool names as keys and pools as values.
        """
        return self._get_pools()

    @abstractmethod
    def _get_pools(self) -> dict[str, type]:
        raise NotImplementedError

    def get_routers(self) -> list[type]:
        """Get the routers value.

        :return: A ``list`` of ``type`` instances.
        """
        return self._get_routers()

    @abstractmethod
    def _get_routers(self) -> list[type]:
        raise NotImplementedError

    def get_middleware(self) -> list[Callable]:
        """Get the middleware value.

        :return: A ``list`` of ``Callable`` instances.
        """
        return self._get_middleware()

    @abstractmethod
    def _get_middleware(self) -> list[type]:
        raise NotImplementedError

    def get_services(self) -> list[type]:
        """Get the services value.

        :return: A ``list`` of ``type`` instances.
        """
        return self._get_services()

    @abstractmethod
    def _get_services(self) -> list[type]:
        raise NotImplementedError

    def get_discovery(self) -> dict[str, Any]:
        """Get the discovery value.

        :return: A ``dict`` instance containing the discovery's config values.
        """
        return self._get_discovery()

    @abstractmethod
    def _get_discovery(self) -> dict[str, Any]:
        raise NotImplementedError

    def get_aggregate(self) -> dict[str, Any]:
        """Get the aggregate value.

        :return: A ``dict`` instance containing the aggregate's config values.
        """
        return self._get_aggregate()

    @abstractmethod
    def _get_aggregate(self) -> dict[str, Any]:
        raise NotImplementedError

    def get_saga(self) -> dict[str, Any]:
        """Get the saga value.

        :return: A ``dict`` instance containing the saga's config values.
        """
        return self._get_saga()

    @abstractmethod
    def _get_saga(self) -> dict[str, Any]:
        raise NotImplementedError

    def get_type_by_key(self, key: str) -> type:
        """Get a type instance by key.

        :param key: The key that identifies the value.
        :return: A ``type`` instance.
        """
        classname = self.get_by_key(key)
        return import_module(classname)

    def get_by_key(self, key: str) -> Any:
        """Get a value by key.

        :param key: The key that identifies the value.
        :return: A value instance.
        """

        def _fn(k: str, data: dict[str, Any], previous: str = "", default: Optional[Any] = sentinel) -> Any:
            current, _sep, following = k.partition(".")
            full = f"{previous}.{current}".lstrip(".")

            with suppress(KeyError):
                return self._parameterized[self._to_parameterized_variable(full)]

            if self._with_environment:
                with suppress(KeyError):
                    return os.environ[self._to_environment_variable(full)]

            if default is not sentinel and current in default:
                default_part = default[current]
            else:
                default_part = sentinel

            if current not in data and default_part is not sentinel:
                part = default_part
            else:
                part = data[current]

            if following:
                return _fn(following, part, full, default_part)

            if not isinstance(part, dict):
                return part

            keys = part.keys()
            if isinstance(default_part, dict):
                keys |= default_part.keys()

            result = dict()
            for subpart in keys:
                result[subpart] = _fn(subpart, part, full, default_part)
            return result

        try:
            return _fn(key, self._data, default=self.DEFAULT_VALUES)
        except Exception:
            raise MinosConfigException(f"{key!r} field is not defined on the configuration!")

    def _to_parameterized_variable(self, key: str) -> str:
        raise KeyError

    def _to_environment_variable(self, key: str) -> str:
        raise KeyError


# noinspection PyUnusedLocal
def _get_version(path: Union[str, Path], *args, **kwargs) -> int:
    if isinstance(path, str):
        path = Path(path)
    if not path.exists():
        raise MinosConfigException(f"Check if this path: {path} is correct")

    with path.open() as file:
        data = yaml.load(file, Loader=yaml.FullLoader)

    return data.get("version", 1)


class MinosConfig(Config, ABC):
    """MinosConfig class."""

    def __new__(cls, *args, **kwargs):
        warnings.warn(f"{MinosConfig!r} has been deprecated. Use {Config} instead.", DeprecationWarning)
        return super().__new__(cls, *args, **kwargs)



from .abc import (
    MinosBinaryProtocol,
)
from .avro import (
    MinosAvroDatabaseProtocol,
    MinosAvroMessageProtocol,
    MinosAvroProtocol,
)
from .json import (
    MinosJsonBinaryProtocol,
)



from __future__ import (
    annotations,
)

from contextlib import (
    suppress,
)
from copy import (
    deepcopy,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Union,
)

from ..exceptions import (
    MinosConfigException,
)
from ..importlib import (
    import_module,
)
from .abc import (
    Config,
)

if TYPE_CHECKING:
    from ..injections import (
        InjectableMixin,
    )


class ConfigV2(Config):
    """Config V2 class."""

    @property
    def _version(self) -> int:
        return 2

    def _get_name(self) -> str:
        return self.get_by_key("name")

    def _get_injections(self) -> list[Union[InjectableMixin, type[InjectableMixin]]]:
        from ..builders import (
            BuildableMixin,
        )
        from ..injections import (
            InjectableMixin,
        )

        partial_ans = list()

        with suppress(MinosConfigException):
            partial_ans.extend(self._get_pools().values())

        with suppress(MinosConfigException):
            partial_ans.append(self._get_interfaces().get("http").get("connector"))

        with suppress(MinosConfigException):
            partial_ans.append(self._get_interfaces().get("broker").get("publisher"))

        with suppress(MinosConfigException):
            partial_ans.append(self._get_interfaces().get("broker").get("subscriber"))

        with suppress(MinosConfigException):
            partial_ans.extend(self._get_aggregate().get("repositories", dict()).values())

        with suppress(MinosConfigException):
            partial_ans.append(self._get_discovery().get("connector"))

        with suppress(MinosConfigException):
            partial_ans.append(self._get_saga().get("manager"))

        with suppress(MinosConfigException):
            injections = self.get_by_key("injections")
            partial_ans.extend(import_module(injection) for injection in injections)

        ans = list()
        for type_ in partial_ans:
            if type_ is None:
                continue

            if isinstance(type_, dict):
                type_ = type_["client"]

            if (
                not issubclass(type_, InjectableMixin)
                and issubclass(type_, BuildableMixin)
                and isinstance((builder_type := type_.get_builder()), InjectableMixin)
            ):
                type_ = builder_type
            elif not issubclass(type_, InjectableMixin):
                raise MinosConfigException(f"{type_!r} must be subclass of {InjectableMixin!r}.")

            ans.append(type_)

        # noinspection PyTypeChecker
        return ans

    def _get_databases(self) -> dict[str, dict[str, Any]]:
        data = deepcopy(self.get_by_key("databases"))
        return data

    def _get_interfaces(self) -> dict[str, dict[str, Any]]:
        data = deepcopy(self.get_by_key("interfaces"))

        if "http" in data:
            data["http"] = self._parse_http_interface(data["http"])
        if "broker" in data:
            data["broker"] = self._parse_broker_interface(data["broker"])
        if "periodic" in data:
            data["periodic"] = self._parse_periodic_interface(data["periodic"])

        return data

    @staticmethod
    def _parse_http_interface(data: dict[str, Any]) -> dict[str, Any]:
        if "port" in data:
            data["port"] = import_module(data["port"])
        if "connector" in data:
            data["connector"]["client"] = import_module(data["connector"]["client"])
        return data

    @staticmethod
    def _parse_broker_interface(data: dict[str, Any]) -> dict[str, Any]:
        if "port" in data:
            data["port"] = import_module(data["port"])

        if "publisher" in data:
            data["publisher"]["client"] = import_module(data["publisher"]["client"])
            if "queue" in data["publisher"]:
                data["publisher"]["queue"] = import_module(data["publisher"]["queue"])

        if "subscriber" in data:
            data["subscriber"]["client"] = import_module(data["subscriber"]["client"])
            if "queue" in data["subscriber"]:
                data["subscriber"]["queue"] = import_module(data["subscriber"]["queue"])
            if "validator" in data["subscriber"]:
                data["subscriber"]["validator"] = import_module(data["subscriber"]["validator"])

        return data

    @staticmethod
    def _parse_periodic_interface(data: dict[str, Any]) -> dict[str, Any]:
        if "port" in data:
            data["port"] = import_module(data["port"])

        return data

    def _get_pools(self) -> dict[str, type]:
        try:
            data = self.get_by_key("pools")
        except MinosConfigException:
            data = dict()

        data = {name: import_module(classname) for name, classname in data.items()}

        return data

    def _get_routers(self) -> list[type]:
        try:
            data = self.get_by_key("routers")
        except MinosConfigException:
            data = list()

        data = [import_module(classname) for classname in data]

        return data

    def _get_middleware(self) -> list[type]:
        try:
            data = self.get_by_key("middleware")
        except MinosConfigException:
            data = list()

        data = [import_module(classname) for classname in data]

        return data

    def _get_services(self) -> list[type]:
        try:
            services = self.get_by_key("services")
        except MinosConfigException:
            services = list()

        services = [import_module(classname) for classname in services]

        return services

    def _get_discovery(self) -> dict[str, Any]:
        data = deepcopy(self.get_by_key("discovery"))
        data["client"] = import_module(data["client"])
        data["connector"] = import_module(data["connector"])
        return data

    def _get_aggregate(self) -> dict[str, Any]:
        data = deepcopy(self.get_by_key("aggregate"))

        data["entities"] = [import_module(classname) for classname in data.get("entities", list())]
        data["repositories"] = {name: import_module(value) for name, value in data.get("repositories", dict()).items()}
        return data

    def _get_saga(self) -> dict[str, Any]:
        data = deepcopy(self.get_by_key("saga"))
        data["manager"] = import_module(data["manager"])
        return data

    def _to_parameterized_variable(self, key: str) -> str:
        return key.replace(".", "_").lower()

    def _to_environment_variable(self, key: str) -> str:
        return f"MINOS_{key.replace('.', '_').upper()}"



import logging
import warnings
from abc import (
    ABC,
)
from asyncio import (
    sleep,
)
from typing import (
    Any,
    Generic,
    Optional,
    TypeVar,
)

from aiomisc import (
    PoolBase,
)
from aiomisc.pool import (
    ContextManager,
)

from .setup import (
    SetupMixin,
)

logger = logging.getLogger(__name__)

P = TypeVar("P")


class Pool(SetupMixin, PoolBase, Generic[P], ABC):
    """Base class for Pool implementations in minos"""

    def __init__(self, *args, maxsize: int = 10, recycle: Optional[int] = 300, already_setup: bool = True, **kwargs):
        super().__init__(*args, maxsize=maxsize, recycle=recycle, already_setup=already_setup, **kwargs)

    # noinspection PyUnresolvedReferences
    async def __acquire(self) -> Any:  # pragma: no cover
        if self._instances.empty() and not self._semaphore.locked():
            await self._PoolBase__create_new_instance()

        instance = await self._instances.get()

        # noinspection PyBroadException
        try:
            result = await self._check_instance(instance)
        except Exception:
            logger.warning("Check instance %r failed", instance)
            self._PoolBase__recycle_instance(instance)
        else:
            if not result:
                self._PoolBase__recycle_instance(instance)
                return await self._PoolBase__acquire()

        self._used.add(instance)
        logger.debug(f"Acquired instance: {instance!r}")
        return instance

    # noinspection PyUnresolvedReferences
    async def __release(self, instance: Any) -> Any:  # pragma: no cover
        await self._PoolBase__release(instance)
        logger.debug(f"Released instance: {instance!r}")

    def acquire(self, *args, **kwargs) -> P:
        """Acquire a new instance wrapped on an asynchronous context manager.

        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: An asynchronous context manager.
        """
        # noinspection PyUnresolvedReferences
        return ContextManager(self.__acquire, self.__release)

    async def _destroy(self) -> None:
        if len(self._used):
            logger.info("Waiting for instances releasing...")
            while len(self._used):
                await sleep(0.1)

        await self.close()

    async def _check_instance(self, instance: P) -> bool:
        return True


class MinosPool(Pool, Generic[P], ABC):
    """MinosPool class."""

    def __init__(self, *args, **kwargs):
        warnings.warn(f"{MinosPool!r} has been deprecated. Use {Pool} instead.", DeprecationWarning)
        super().__init__(*args, **kwargs)



from .abc import (
    DataDecoder,
    DataEncoder,
    SchemaDecoder,
    SchemaEncoder,
)
from .avro import (
    AvroDataDecoder,
    AvroDataEncoder,
    AvroSchemaDecoder,
    AvroSchemaEncoder,
)



from __future__ import (
    annotations,
)

from pathlib import (
    Path,
)
from typing import (
    Any,
    Optional,
    Type,
    Union,
)

import lmdb

from ..protocol import (
    MinosAvroDatabaseProtocol,
    MinosBinaryProtocol,
)
from .abc import (
    MinosStorage,
)


class MinosStorageLmdb(MinosStorage):
    """Minos Storage LMDB class"""

    __slots__ = "_env", "_protocol", "_tables"

    # noinspection PyUnusedLocal
    def __init__(
        self, env: lmdb.Environment, protocol: Type[MinosBinaryProtocol] = MinosAvroDatabaseProtocol, **kwargs
    ):
        self._env: lmdb.Environment = env
        self._protocol = protocol
        self._tables = {}

    def add(self, table: str, key: str, value: Any) -> None:
        """Store a value.

        :param table: Table in which the data is stored.
        :param key: Key that identifies the data.
        :param value: Data to be stored.
        :return: This method does not return anything.
        """
        db_instance = self._get_table(table)
        with self._env.begin(write=True) as txn:
            value_bytes: bytes = self._protocol.encode(value)
            txn.put(key.encode(), value_bytes, db=db_instance)

    def get(self, table: str, key: str) -> Optional[Any]:
        """Get the stored value..

        :param table: Table in which the data is stored.
        :param key: Key that identifies the data.
        :return: The stored value or ``None`` if it's empty.
        """
        db_instance = self._get_table(table)
        with self._env.begin(db=db_instance) as txn:
            value_binary = txn.get(key.encode())
            if value_binary is not None:
                # decode the returned value
                return self._protocol.decode(value_binary)
            return None

    def delete(self, table: str, key: str) -> None:
        """Delete the stored value.

        :param table: Table in which the data is stored.
        :param key: Key that identifies the data.
        :return: This method does not return anything.
        """
        db_instance = self._get_table(table)
        with self._env.begin(write=True, db=db_instance) as txn:
            txn.delete(key.encode())

    def update(self, table: str, key: str, value: Any) -> None:
        """Update the stored value.

        :param table: Table in which the data is stored.
        :param key: Key that identifies the data.
        :param value: Data to be stored.
        :return: This method does not return anything.
        """
        db_instance = self._get_table(table)
        with self._env.begin(write=True, db=db_instance) as txn:
            value_bytes: bytes = self._protocol.encode(value)
            txn.put(key.encode(), value_bytes, db=db_instance, overwrite=True)

    def _get_table(self, table: str):
        if table in self._tables:
            return self._tables[table]
        else:
            # create a new table
            self._tables[table] = self._env.open_db(table.encode())
            return self._tables[table]

    @classmethod
    def build(
        cls, path: Optional[Union[str, Path]] = None, max_db: int = 10, map_size: int = int(1e9), **kwargs
    ) -> MinosStorageLmdb:
        """Build a new instance.

        :param path: Path in which the database is stored.
        :param max_db: Maximum number of available databases.
        :param map_size: Maximum number of entries to be stored on the database. Default set to 1GB
        :param kwargs: Additional named arguments.
        :return: A ``MinosStorageLmdb`` instance.
        """
        if path is None:
            path = ".lmdb"

        env: lmdb.Environment = lmdb.open(str(path), max_dbs=max_db, map_size=map_size)
        return cls(env, **kwargs)



from __future__ import (
    annotations,
)

import types
from asyncio import (
    iscoroutinefunction,
)
from contextlib import (
    suppress,
)
from functools import (
    wraps,
)
from inspect import (
    Parameter,
    signature,
)
from typing import (
    Any,
    Generic,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from dependency_injector.containers import (
    Container,
)
from dependency_injector.wiring import (
    Provide,
    inject,
)

from ..exceptions import (
    NotProvidedException,
)
from ..model.types import (
    is_type_subclass,
)
from .mixins import (
    InjectableMixin,
)

InputType = TypeVar("InputType", bound=type)

OutputType = Union[InputType, Type[InjectableMixin]]


class Injectable:
    """Injectable decorator."""

    def __init__(self, name: str):
        if not isinstance(name, str):
            raise ValueError(f"'name' argument must be a {str!r} instance. Obtained: {name}")
        self._name = name

    @property
    def name(self) -> str:
        """Get the name.

        :return: A ``str`` value.
        """
        return self._name

    def __call__(self, input_type: InputType) -> OutputType:
        bases = (input_type, InjectableMixin)
        if (generic := self._build_generic(input_type)) is not None:
            bases = (*bases, generic)

        # noinspection PyTypeChecker
        output_type: OutputType = types.new_class(input_type.__name__, bases, {})
        output_type.__module__ = input_type.__module__

        # noinspection PyProtectedMember
        output_type._set_injectable_name(self._name)

        return output_type

    @staticmethod
    def _build_generic(type_):
        generic = next(
            (base for base in getattr(type_, "__orig_bases__", tuple()) if get_origin(base) is not None), None
        )
        if generic is None:
            return None

        generics = tuple(arg for arg in get_args(generic) if isinstance(arg, TypeVar))
        if not len(generics):
            return None

        # noinspection PyTypeHints
        return Generic[generics]


class Inject:
    """Inject class."""

    def __call__(self, func):
        type_hints_ = self._build_type_hints(func)

        if iscoroutinefunction(func):

            @wraps(func)
            async def _wrapper(*args, **kwargs):
                kwargs = self._inject_on_kwargs(type_hints_, args, kwargs)
                return await func(*args, **kwargs)

        else:

            @wraps(func)
            def _wrapper(*args, **kwargs):
                kwargs = self._inject_on_kwargs(type_hints_, args, kwargs)
                return func(*args, **kwargs)

        return _wrapper

    def _inject_on_kwargs(
        self, type_hints_: dict[str, tuple[int, type[V], bool]], args: tuple[Any], kwargs: dict[str, Any]
    ) -> dict[str, Any]:
        for name in type_hints_.keys() - kwargs.keys():
            if type_hints_[name][0] < len(args):
                continue
            try:
                kwargs[name] = self.resolve(type_hints_[name][1])
            except NotProvidedException as exc:
                if not type_hints_[name][2]:
                    raise exc
        return kwargs

    def _build_type_hints(self, func) -> dict[str, tuple[int, type[V], bool]]:
        type_hints_ = dict()

        hints = get_type_hints(func)

        for i, (name, field) in enumerate(signature(func).parameters.items()):
            if name in ("return", "self", "cls") or name not in hints:
                continue
            type_ = hints[name]
            if not self._is_injectable(type_):
                continue
            type_hints_[name] = (i, type_, field.default is not Parameter.empty)
        return type_hints_

    def _is_injectable(self, type_: type) -> bool:
        origin_type = get_origin(type_)
        if origin_type is Union:
            if any(self._is_injectable(arg) for arg in get_args(type_)):
                return True
        elif origin_type is None:
            if issubclass(type_, InjectableMixin):
                return True
        elif is_type_subclass(origin_type):
            if issubclass(origin_type, InjectableMixin):
                return True
        return False

    @classmethod
    def resolve(cls, type_: type[V]) -> V:
        """Resolve a dependency by type.

        :param type_: The type of the dependency.
        :return: The dependency value.
        """

        origin_type = get_origin(type_)

        if origin_type is Union:
            for arg in get_args(type_):
                with suppress(NotProvidedException):
                    return cls.resolve(arg)
        elif origin_type is None:
            if issubclass(type_, InjectableMixin):
                return cls.resolve_by_name(type_.get_injectable_name())
        elif is_type_subclass(origin_type):
            if issubclass(origin_type, InjectableMixin):
                return cls.resolve_by_name(origin_type.get_injectable_name())

        raise NotProvidedException(f"The given type injection could not be resolved: {type_}")

    @staticmethod
    @inject
    def resolve_by_name(name: str, container: Container = Provide["<container>"]) -> Any:
        """Resolve a dependency by name.

        :param name: The name of the dependency.
        :param container: The container that contains the dependency.
        :return: The dependency value.
        """
        try:
            return container.providers[name]()
        except Exception:
            raise NotProvidedException(f"The {name} injection must be provided.")


V = TypeVar("V")



from __future__ import (
    annotations,
)

import logging
from functools import (
    lru_cache,
)
from typing import (
    Any,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

from .generics import (
    unpack_typevar,
)
from .model_types import (
    ModelType,
)


def is_model_subclass(type_: type) -> bool:
    """Check if the given type field is subclass of ``Model``."""
    from ..abc import (
        Model,
    )

    if not is_type_subclass(type_):
        type_ = get_origin(type_)
    return is_type_subclass(type_) and issubclass(type_, Model)


def is_type_subclass(type_: type) -> bool:
    """Check if the given type field is subclass of ``type``."""
    return issubclass(type(type_), type(type))


def is_model_type(value_: type):
    """Check if the given type is a model instance."""
    from ..abc import (
        Model,
    )

    return isinstance(value_, Model)


logger = logging.getLogger(__name__)

T = TypeVar("T", bound=type)
K = TypeVar("K", bound=type)


class TypeHintComparator:
    """Type Hint Comparator class."""

    def __init__(self, first: T, second: K):
        self._first = first
        self._second = second

    def match(self) -> bool:
        """Check if the two types match.

        :return: ``True`` if there is a match or ``False`` otherwise.
        """

        return self._compare(self._first, self._second)

    @classmethod
    @lru_cache()
    def _compare(cls, first: T, second: K) -> bool:
        if isinstance(first, TypeVar):
            first = unpack_typevar(first)

        if isinstance(second, TypeVar):
            second = unpack_typevar(second)

        if second is Any:
            return True

        if get_origin(first) is Union and all(cls._compare(f, second) for f in get_args(first)):
            return True

        if get_origin(second) is Union and any(cls._compare(first, s) for s in get_args(second)):
            return True

        if is_model_subclass(first):
            first = ModelType.from_model(first)

        if is_model_subclass(second):
            second = ModelType.from_model(second)

        if first == second:
            return True

        if is_type_subclass(first) and is_type_subclass(second):
            if isinstance(first, ModelType) and isinstance(second, ModelType):
                if first <= second:
                    return True
            elif issubclass(first, second):
                return True

        first_origin, second_origin = get_origin(first), get_origin(second)
        if first_origin is not None and cls._compare(first_origin, second_origin):
            return cls._compare_args(first, second)

        return False

    @classmethod
    def _compare_args(cls, first: T, second: K) -> bool:
        first_args, second_args = get_args(first), get_args(second)
        if len(first_args) != len(second_args):
            return False
        return all(cls._compare(fi, si) for fi, si in zip(first_args, second_args))



from .abc import (
    Model,
)
from .declarative import (
    DeclarativeModel,
    MinosModel,
)
from .dynamic import (
    BucketModel,
    DataTransferObject,
    DynamicModel,
)
from .fields import (
    Field,
    ModelField,
)
from .serializers import (
    AvroDataDecoder,
    AvroDataEncoder,
    AvroSchemaDecoder,
    AvroSchemaEncoder,
    DataDecoder,
    DataEncoder,
    SchemaDecoder,
    SchemaEncoder,
)
from .types import (
    GenericTypeProjector,
    MissingSentinel,
    ModelType,
    NoneType,
    TypeHintBuilder,
    TypeHintComparator,
    is_model_type,
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['minos',
 'minos.common',
 'minos.common.config',
 'minos.common.database',
 'minos.common.injections',
 'minos.common.model',
 'minos.common.model.dynamic',
 'minos.common.model.serializers',
 'minos.common.model.serializers.avro',
 'minos.common.model.serializers.avro.data',
 'minos.common.model.serializers.avro.schema',
 'minos.common.model.types',
 'minos.common.protocol',
 'minos.common.protocol.avro',
 'minos.common.storage']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<7.0.0',
 'aiomisc>=14.0.3,<15.8.0',
 'aiopg>=1.2.1,<2.0.0',
 'cached-property>=1.5.2,<2.0.0',
 'dependency-injector>=4.32.2,<5.0.0',
 'fastavro>=1.4.0,<2.0.0',
 'lmdb>=1.2.1,<2.0.0',
 'orjson>=3.5.2,<4.0.0',
 'uvloop>=0.16.0,<0.17.0']

setup_kwargs = {
    'name': 'minos-microservice-common',
    'version': '0.7.0.dev1',
    'description': 'The common core of the Minos Framework',
    'long_description': '<p align="center">\n  <a href="https://minos.run" target="_blank"><img src="https://raw.githubusercontent.com/minos-framework/.github/main/images/logo.png" alt="Minos logo"></a>\n</p>\n\n## minos-microservice-common\n\n[![PyPI Latest Release](https://img.shields.io/pypi/v/minos-microservice-common.svg)](https://pypi.org/project/minos-microservice-common/)\n[![GitHub Workflow Status](https://img.shields.io/github/workflow/status/minos-framework/minos-python/pages%20build%20and%20deployment?label=docs)](https://minos-framework.github.io/minos-python)\n[![License](https://img.shields.io/github/license/minos-framework/minos-python.svg)](https://github.com/minos-framework/minos-python/blob/main/LICENSE)\n[![Coverage](https://codecov.io/github/minos-framework/minos-python/coverage.svg?branch=main)](https://codecov.io/gh/minos-framework/minos-python)\n[![Stack Overflow](https://img.shields.io/badge/Stack%20Overflow-Ask%20a%20question-green)](https://stackoverflow.com/questions/tagged/minos)\n\n## Summary\n\nMinos is a framework which helps you create [reactive](https://www.reactivemanifesto.org/) microservices in Python.\nInternally, it leverages Event Sourcing, CQRS and a message driven architecture to fulfil the commitments of an\nasynchronous environment.\n\n## Documentation\n\nThe official API Reference is publicly available at the [GitHub Pages](https://minos-framework.github.io/minos-python).\n\n## Source Code\n\nThe source code of this project is hosted at the [GitHub Repository](https://github.com/minos-framework/minos-python).\n\n## Getting Help\n\nFor usage questions, the best place to go to is [StackOverflow](https://stackoverflow.com/questions/tagged/minos).\n\n## Discussion and Development\nMost development discussions take place over the [GitHub Issues](https://github.com/minos-framework/minos-python/issues). In addition, a [Gitter channel](https://gitter.im/minos-framework/community) is available for development-related questions.\n\n## License\n\nThis project is distributed under the [MIT](https://raw.githubusercontent.com/minos-framework/minos-python/main/LICENSE) license.\n',
    'author': 'Minos Framework Devs',
    'author_email': 'hey@minos.run',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.minos.run',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from collections.abc import (
    Hashable,
)
from typing import (
    AsyncContextManager,
    Optional,
)

from aiopg import (
    Connection,
    Cursor,
)

from ..locks import (
    Lock,
)


class PostgreSqlLock(Lock):
    """ "PostgreSql Lock class."""

    cursor: Optional[Cursor]

    def __init__(self, wrapped_connection: AsyncContextManager[Connection], key: Hashable, *args, **kwargs):
        super().__init__(key, *args, **kwargs)

        self.wrapped_connection = wrapped_connection
        self.cursor = None

        self._args = args
        self._kwargs = kwargs

    async def __aenter__(self):
        connection = await self.wrapped_connection.__aenter__()
        cursor = await connection.cursor(*self._args, **self._kwargs).__aenter__()

        self.cursor = cursor
        await self.cursor.execute("select pg_advisory_lock(%(hashed_key)s)", {"hashed_key": self.hashed_key})
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.cursor.execute("select pg_advisory_unlock(%(hashed_key)s)", {"hashed_key": self.hashed_key})
        if not self.cursor.closed:
            self.cursor.close()
        self.cursor = None
        await self.wrapped_connection.__aexit__(exc_type, exc_val, exc_tb)



import logging
from asyncio import (
    sleep,
)
from collections.abc import (
    Awaitable,
    Callable,
    Iterable,
)
from inspect import (
    isawaitable,
)
from typing import (
    Generic,
    TypeVar,
    Union,
)

from aiomisc import (
    CircuitBreaker,
    CircuitBroken,
)
from aiomisc.circuit_breaker import (
    CircuitBreakerStates,
)

from .object import (
    Object,
)

logger = logging.getLogger(__name__)
R = TypeVar("R")
E = TypeVar("E")


class CircuitBreakerMixin(Generic[E], Object):
    """Circuit Breaker Mixin class."""

    def __init__(
        self,
        circuit_breaker_error_rate: float = 0.2,
        circuit_breaker_time: Union[int, float] = 3,
        circuit_breaker_exceptions: Iterable[type[Exception]] = (Exception,),
        **kwargs,
    ):
        super().__init__(**kwargs)

        self._circuit_breaker_error_rate = circuit_breaker_error_rate
        self._circuit_breaker_time = circuit_breaker_time
        self._circuit_breaker_exceptions = tuple(circuit_breaker_exceptions)

        self._circuit_breaker = CircuitBreaker(
            error_ratio=self._circuit_breaker_error_rate,
            response_time=self._circuit_breaker_time,
            exceptions=self._circuit_breaker_exceptions,
        )

    async def with_circuit_breaker(self, fn: Callable[[], Union[Awaitable[R], R]]) -> R:
        """Execute the given function with circuit breaker.

        :param fn: The function to be executed.
        :return: The return of the given function.
        """
        while True:
            try:
                with self._circuit_breaker.context():
                    ans = fn()
                    if isawaitable(ans):
                        ans = await ans
                    return ans
            except CircuitBroken:
                await sleep(self._circuit_breaker_broken_timeout)
            except self._circuit_breaker_exceptions as exc:
                logger.warning(f"An exception was raised: {exc!r}")
                await sleep(self._circuit_breaker_exception_timeout)

    @property
    def _circuit_breaker_broken_timeout(self) -> float:
        return self._circuit_breaker_time * 0.5

    @property
    def _circuit_breaker_exception_timeout(self) -> float:
        return self._circuit_breaker_time * 0.05

    @property
    def circuit_breaker(self) -> CircuitBreaker:
        """Get the circuit breaker.

        :return: A ``CircuitBreaker`` instance.
        """
        return self._circuit_breaker

    @property
    def is_circuit_breaker_passing(self) -> bool:
        """Check if circuit breaker is passing.

        :return: A ``bool`` instance.
        """
        return self._circuit_breaker.state == CircuitBreakerStates.PASSING

    @property
    def is_circuit_breaker_broken(self) -> bool:
        """Check if circuit breaker is passing.

        :return: A ``bool`` instance.
        """
        return self._circuit_breaker.state == CircuitBreakerStates.BROKEN

    @property
    def is_circuit_breaker_recovering(self) -> bool:
        """Check if circuit breaker is passing.

        :return: A ``bool`` instance.
        """
        return self._circuit_breaker.state == CircuitBreakerStates.RECOVERING

    @property
    def circuit_breaker_exceptions(self) -> tuple[type[Exception]]:
        """Get the circuit breaker exceptions.

        :return: A tuple of ``Exception`` types.
        """
        return self._circuit_breaker_exceptions



from __future__ import (
    annotations,
)

from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Union,
)

if TYPE_CHECKING:
    from ..abc import (
        Model,
    )
    from ..fields import (
        Field,
    )


class SchemaEncoder(ABC):
    """Schema Encoder base class."""

    @abstractmethod
    def build(self, type_: Union[type, Model, Field], **kwargs) -> Any:
        """Build the avro schema for the given field.

        :param type_: The type to be encoded as a schema.
        :return: A dictionary object.
        """


class SchemaDecoder(ABC):
    """Schema Decoder base class."""

    @abstractmethod
    def build(self, schema: Any, **kwargs) -> type:
        """Build type from given avro schema item.

        :param schema: The schema to be decoded as a type.
        :return: A type object.
        """


class DataEncoder(ABC):
    """Data Encoder base class."""

    @abstractmethod
    def build(self, value: Any, **kwargs) -> Any:
        """Build an avro data representation based on the content of the given field.

        :param value: The value to be encoded.
        :return: A `avro`-compatible data.
        """


class DataDecoder(ABC):
    """Data Decoder base class."""

    @abstractmethod
    def build(self, data: Any, type_: Any, **kwargs) -> Any:
        """Cast data type according to the field definition.

        :param data: Data to be casted.
        :param type_: The type of the decoded data.
        :return: The casted object.
        """



from __future__ import (
    annotations,
)

from asyncio import (
    gather,
)
from typing import (
    TYPE_CHECKING,
    Optional,
    Type,
    Union,
)

from cached_property import (
    cached_property,
)
from dependency_injector import (
    containers,
    providers,
)

from ..importlib import (
    import_module,
)
from .mixins import (
    InjectableMixin,
)

if TYPE_CHECKING:
    from ..config import (
        Config,
    )


class DependencyInjector:
    """Async wrapper of ``dependency_injector.containers.Container``."""

    def __init__(
        self,
        config: Config,
        injections: Optional[list[Union[InjectableMixin, Type[InjectableMixin], str]]] = None,
    ):
        if injections is None:
            injections = list()
        self.config = config
        self._raw_injections = injections

    @cached_property
    def injections(self) -> dict[str, InjectableMixin]:
        """Get the injections' dictionary.

        :return: A dict of injections.
        """
        injections = dict()

        def _fn(raw: Union[InjectableMixin, type[InjectableMixin], str]) -> InjectableMixin:
            if isinstance(raw, str):
                raw = import_module(raw)
            if isinstance(raw, type):
                return raw.from_config(self.config, **injections)
            return raw

        for value in self._raw_injections:
            try:
                value = _fn(value)
                key = value.get_injectable_name()
            except Exception as exc:
                raise ValueError(f"An exception was raised while building injections: {exc!r}")
            injections[key] = value

        return injections

    async def wire_and_setup_injections(self, *args, **kwargs) -> None:
        """Connect the configuration.

        :return: This method does not return anything.
        """

        self.wire_injections(*args, **kwargs)
        await self.setup_injections()

    async def unwire_and_destroy_injections(self) -> None:
        """Disconnect the configuration.

        :return: This method does not return anything.
        """
        await self.destroy_injections()
        self.unwire_injections()

    def wire_injections(self, modules=None, *args, **kwargs) -> None:
        """Connect the configuration.

        :return: This method does not return anything.
        """
        from . import (
            decorators,
        )

        if modules is None:
            modules = tuple()
        modules = [*modules, decorators]

        self.container.wire(modules=modules, *args, **kwargs)

    def unwire_injections(self) -> None:
        """Disconnect the configuration.

        :return: This method does not return anything.
        """
        self.container.unwire()

    async def setup_injections(self) -> None:
        """Set Up the injections.

        :return: This method does not return anything.
        """
        await gather(*(injection.setup() for injection in self.injections.values()))

    async def destroy_injections(self) -> None:
        """Destroy the injections

        :return: This method does not return anything.
        """
        await gather(*(injection.destroy() for injection in self.injections.values()))

    @cached_property
    def container(self) -> containers.Container:
        """Get the dependencies' container.

        :return: A ``Container`` instance.
        """
        container = containers.DynamicContainer()
        container.config = providers.Object(self.config)

        for name, injection in self.injections.items():
            container.set_provider(name, providers.Object(injection))
        return container

    def __getattr__(self, item: str) -> InjectableMixin:
        if item not in self.injections:
            raise AttributeError(f"{type(self).__name__!r} does not contain the {item!r} attribute.")
        return self.injections[item]



import unittest
from itertools import (
    starmap,
)
from pathlib import (
    Path,
)
from typing import (
    Any,
)
from uuid import (
    uuid4,
)

import aiopg

from .config import (
    Config,
)


class PostgresAsyncTestCase(unittest.IsolatedAsyncioTestCase):
    CONFIG_FILE_PATH: Path

    def setUp(self) -> None:
        self._uuid = uuid4()
        self._config = Config(self.CONFIG_FILE_PATH)

        self._meta_repository_db = self._config.get_database_by_name("aggregate")

        self._meta_broker_queue_db = self._config.get_database_by_name("broker")

        self._meta_snapshot_db = self._config.get_database_by_name("aggregate")

        self._test_db = {"database": f"test_db_{self._uuid.hex}", "user": f"test_user_{self._uuid.hex}"}

        self.repository_db = self._meta_repository_db | self._test_db
        self.broker_queue_db = self._meta_broker_queue_db | self._test_db
        self.snapshot_db = self._meta_snapshot_db | self._test_db

        self.config = Config(
            self.CONFIG_FILE_PATH,
            repository_database=self.repository_db["database"],
            repository_user=self.repository_db["user"],
            broker_queue_database=self.broker_queue_db["database"],
            broker_queue_user=self.broker_queue_db["user"],
            snapshot_database=self.snapshot_db["database"],
            snapshot_user=self.snapshot_db["user"],
        )

    async def asyncSetUp(self):
        pairs = self._drop_duplicates(
            [
                (self._meta_repository_db, self.repository_db),
                (self._meta_broker_queue_db, self.broker_queue_db),
                (self._meta_snapshot_db, self.snapshot_db),
            ]
        )
        for meta, test in pairs:
            await self._setup_database(dict(meta), dict(test))

    async def _setup_database(self, meta: dict[str, Any], test: dict[str, Any]) -> None:
        await self._teardown_database(meta, test)

        async with aiopg.connect(**meta) as connection:
            async with connection.cursor() as cursor:
                template = "CREATE ROLE {user} WITH SUPERUSER CREATEDB LOGIN ENCRYPTED PASSWORD {password!r};"
                await cursor.execute(template.format(**test))

                template = "CREATE DATABASE {database} WITH OWNER = {user};"
                await cursor.execute(template.format(**test))

    async def asyncTearDown(self):
        pairs = self._drop_duplicates(
            [
                (self._meta_repository_db, self.repository_db),
                (self._meta_broker_queue_db, self.broker_queue_db),
                (self._meta_snapshot_db, self.snapshot_db),
            ]
        )

        for meta, test in pairs:
            await self._teardown_database(meta, test)

    @staticmethod
    def _drop_duplicates(items: list[(dict, dict)]):
        items = starmap(lambda a, b: (tuple(a.items()), tuple(b.items())), items)
        items = set(items)
        items = starmap(lambda a, b: (dict(a), dict(b)), items)
        items = list(items)
        return items

    @staticmethod
    async def _teardown_database(meta: dict[str, Any], test: dict[str, Any]) -> None:
        async with aiopg.connect(**meta) as connection:
            async with connection.cursor() as cursor:
                template = "DROP DATABASE IF EXISTS {database}"
                await cursor.execute(template.format(**test))

                template = "DROP ROLE IF EXISTS {user};"
                await cursor.execute(template.format(**test))



import logging
from asyncio import (
    sleep,
)
from collections.abc import (
    Hashable,
)
from typing import (
    Optional,
)

import aiopg
from aiomisc.pool import (
    ContextManager,
)
from aiopg import (
    Connection,
)
from psycopg2 import (
    OperationalError,
)

from ..injections import (
    Injectable,
)
from ..locks import (
    LockPool,
)
from ..pools import (
    Pool,
)
from .locks import (
    PostgreSqlLock,
)

logger = logging.getLogger(__name__)


@Injectable("postgresql_pool")
class PostgreSqlPool(Pool[ContextManager]):
    """Postgres Pool class."""

    def __init__(
        self,
        database: str,
        host: Optional[str] = None,
        port: Optional[int] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if host is None:
            host = "localhost"
        if port is None:
            port = 5432
        if user is None:
            user = "postgres"
        if password is None:
            password = ""

        self.database = database
        self.host = host
        self.port = port
        self.user = user
        self.password = password

    @classmethod
    def _from_config(cls, *args, config, **kwargs):
        return cls(*args, **config.get_default_database(), **kwargs)

    async def _create_instance(self) -> Optional[Connection]:
        try:
            connection = await aiopg.connect(
                host=self.host, port=self.port, dbname=self.database, user=self.user, password=self.password
            )
        except OperationalError as exc:
            logger.warning(f"There was an {exc!r} while trying to get a database connection.")
            await sleep(1)
            return None

        logger.info(f"Created {self.database!r} database connection identified by {id(connection)}!")
        return connection

    async def _destroy_instance(self, instance: Connection):
        if not instance.closed:
            await instance.close()
        logger.info(f"Destroyed {self.database!r} database connection identified by {id(instance)}!")

    async def _check_instance(self, instance: Optional[Connection]) -> bool:
        if instance is None:
            return False

        try:
            # This operation connects to the database and raises an exception if something goes wrong.
            instance.isolation_level
        except OperationalError:
            return False

        return not instance.closed


class PostgreSqlLockPool(LockPool, PostgreSqlPool):
    """Postgres Locking Pool class."""

    def acquire(self, key: Hashable, *args, **kwargs) -> PostgreSqlLock:
        """Acquire a new lock.

        :param key: The key to be used for locking.
        :return: A ``PostgreSqlLock`` instance.
        """
        return PostgreSqlLock(super().acquire(), key, *args, **kwargs)



AVRO_BOOLEAN = "boolean"
AVRO_NULL = "null"
AVRO_INT = "int"
AVRO_FLOAT = "float"
AVRO_LONG = "long"
AVRO_DOUBLE = "double"
AVRO_BYTES = "bytes"
AVRO_STRING = "string"
AVRO_ARRAY = "array"
AVRO_RECORD = "record"
AVRO_ENUM = "enum"
AVRO_MAP = "map"
AVRO_FIXED = "fixed"
AVRO_DECIMAL = "decimal"

AVRO_DATE = {"type": AVRO_INT, "logicalType": "date"}
AVRO_TIME = {"type": AVRO_INT, "logicalType": "time-micros"}
AVRO_TIMESTAMP = {"type": AVRO_LONG, "logicalType": "timestamp-micros"}
AVRO_TIMEDELTA = {"type": AVRO_LONG, "logicalType": "timedelta-micros"}
AVRO_UUID = {"type": AVRO_STRING, "logicalType": "uuid"}
AVRO_SET = {"type": AVRO_ARRAY, "logicalType": "set"}



from __future__ import (
    annotations,
)

import logging
from contextlib import (
    suppress,
)
from datetime import (
    date,
    datetime,
    time,
    timedelta,
)
from typing import (
    Any,
    Union,
)
from uuid import (
    UUID,
)

from .....exceptions import (
    MinosImportException,
    MinosMalformedAttributeException,
)
from .....importlib import (
    import_module,
)
from ....types import (
    MissingSentinel,
    ModelType,
    NoneType,
    build_union,
    is_model_subclass,
)
from ...abc import (
    SchemaDecoder,
)
from .constants import (
    AVRO_ARRAY,
    AVRO_BOOLEAN,
    AVRO_BYTES,
    AVRO_DATE,
    AVRO_DOUBLE,
    AVRO_FLOAT,
    AVRO_INT,
    AVRO_MAP,
    AVRO_NULL,
    AVRO_RECORD,
    AVRO_SET,
    AVRO_STRING,
    AVRO_TIME,
    AVRO_TIMEDELTA,
    AVRO_TIMESTAMP,
    AVRO_UUID,
)

logger = logging.getLogger(__name__)


class AvroSchemaDecoder(SchemaDecoder):
    """Avro Schema Decoder class."""

    def __init__(self, schema: Any = None):
        self._schema = schema

    def build(self, schema: Any = MissingSentinel, **kwargs) -> type:
        """Build type from given avro schema item.

        :param schema: The schema to be decoded as a type.
        :return: A type object.
        """
        if schema is MissingSentinel:
            schema = self._schema
        return self._build(schema, **kwargs)

    def _build(self, schema: Union[dict, list, str], **kwargs) -> type:
        if isinstance(schema, dict):
            return self._build_from_dict(schema, **kwargs)

        if isinstance(schema, list):
            return self._build_from_list(schema, **kwargs)

        return self._build_simple(schema, **kwargs)

    def _build_from_list(self, schema: list[Any], **kwargs) -> type:
        options = list()
        for entry in schema:
            with suppress(Exception):
                options.append(self._build(entry, **kwargs))
        return build_union(options)

    def _build_from_dict(self, schema: dict, **kwargs) -> type:
        if "logicalType" in schema:
            return self._build_logical_type(schema, **kwargs)

        if schema["type"] == AVRO_ARRAY:
            return self._build_list(schema, **kwargs)

        if schema["type"] == AVRO_MAP:
            return self._build_dict(schema, **kwargs)

        if schema["type"] == AVRO_RECORD:
            return self._build_record(schema, **kwargs)

        return self._build_type(schema, **kwargs)

    def _build_logical_type(self, schema: dict[str, Any], **kwargs) -> type:
        logical_type = schema["logicalType"]
        if logical_type == AVRO_DATE["logicalType"]:
            return date

        if logical_type == AVRO_TIME["logicalType"]:
            return time

        if logical_type == AVRO_TIMESTAMP["logicalType"]:
            return datetime

        if logical_type == AVRO_TIMEDELTA["logicalType"]:
            return timedelta

        if logical_type == AVRO_UUID["logicalType"]:
            return UUID

        if logical_type == AVRO_SET["logicalType"]:
            return self._build_set(schema, **kwargs)

        sub_schema = {k: v for k, v in schema.items() if k != "logicalType"}

        try:
            cls_ = import_module(logical_type)
        except MinosImportException:
            cls_ = None

        if cls_ is not None:
            # noinspection PyUnresolvedReferences
            if is_model_subclass(cls_) and (ans := cls_.decode_schema(self, sub_schema)) is not MissingSentinel:
                return ans

            return cls_

        return self._build(sub_schema, **kwargs)

    def _build_list(self, schema: dict[str, Any], **kwargs) -> type:
        items = schema["items"]
        return list[self._build_iterable(items, **kwargs)]

    def _build_set(self, schema: dict[str, Any], **kwargs) -> type:
        items = schema["items"]
        return set[self._build_iterable(items, **kwargs)]

    def _build_dict(self, schema: dict[str, Any], **kwargs) -> type:
        values = schema["values"]
        return dict[str, self._build_iterable(values, **kwargs)]

    def _build_iterable(self, values: Union[dict, str, Any], **kwargs) -> type:
        type_ = self._build(values, **kwargs)
        if type_ is NoneType:
            # FIXME: This is a design decision that must be revisited in the future.
            type_ = Any
        return type_

    def _build_record(self, schema: dict[str, Any], **kwargs) -> type:
        name, namespace = schema["name"], schema.get("namespace")
        if namespace is None:
            try:
                namespace, name = name.rsplit(".", 1)
            except ValueError:
                namespace = str()

        namespace = self._unpatch_namespace(namespace)

        if len(namespace) > 0:
            classname = f"{namespace}.{name}"
        else:
            classname = name

        try:
            cls_ = import_module(classname)
        except MinosImportException:
            cls_ = None

        if is_model_subclass(cls_) and (ans := cls_.decode_schema(self, schema, **kwargs)) is not MissingSentinel:
            return ans

        type_hints = {field["name"]: self._build(field, **kwargs) for field in schema["fields"]}

        return ModelType.build(name_=name, type_hints_=type_hints, namespace_=namespace)

    @staticmethod
    def _unpatch_namespace(namespace: str) -> str:
        if len(namespace) > 0:
            return namespace.rsplit(".", 1)[0]
        return namespace

    def _build_type(self, schema: dict[str, Any], **kwargs) -> type:
        return self._build(schema["type"], **kwargs)

    @staticmethod
    def _build_simple(type_: str, **kwargs) -> type:
        if type_ == AVRO_NULL:
            return NoneType

        if type_ == AVRO_INT:
            return int

        if type_ == AVRO_BOOLEAN:
            return bool

        if type_ == AVRO_FLOAT:
            return float

        if type_ == AVRO_DOUBLE:
            return float

        if type_ == AVRO_STRING:
            return str

        if type_ == AVRO_BYTES:
            return bytes

        raise MinosMalformedAttributeException(f"Given field type is not supported: {type_!r}")
